import numpy as np
import textwrap
import copy
import collections
import warnings
import random
from scipy.stats import norm, beta
import networkx as nx
from networkx.algorithms.flow import shortest_augmenting_path

#from mbnpy.utils import all_equal
from mbnpy import variable
from mbnpy import utils
from mbnpy import cpm


def condition(cpms, cnd_vars, cnd_states):
    """
    Returns a list of instance of Cpm

    Parameters
    ----------
    cpms: a dict or list of instances of Cpm
    cnd_vars: a list of variables to be conditioned
    cnd_states: a list of the states to be conditioned
    """

    assert isinstance(cpms, (cpm.Cpm, list, dict)), 'invalid cpms'

    keys = None
    if isinstance(cpms, cpm.Cpm):
        cpms = [cpms]
    elif isinstance(cpms, dict):
        keys = list(cpms.keys())
        cpms = list(cpms.values())

    assert isinstance(cnd_vars, (list, np.ndarray)), 'invalid cnd_vars'

    if isinstance(cnd_vars, np.ndarray):
        cnd_vars = cnd_vars.tolist()

    if cnd_vars and isinstance(cnd_vars[0], str):
        cnd_vars = get_variables(cpms, cnd_vars)

    assert isinstance(cnd_states, (list, np.ndarray)), 'invalid cnd_vars'

    if isinstance(cnd_states, np.ndarray):
        cnd_states = cnd_states.tolist()

    if cnd_states and isinstance(cnd_states[0], str):
        cnd_states = [x.values.index(y) for x, y in zip(cnd_vars, cnd_states)]

    Mc = [Mx.condition(cnd_vars, cnd_states) for Mx in cpms]

    if keys:
        return {k: M for k, M in zip(keys, Mc)}
    else:
        return Mc


def get_var_idx(varis, names):
    """
    INPUT:
    varis: a list of variables
    names: a list of variable names
    OUTPUT:
    idx: a list of column indices of v_names
    """

    idx = []
    for v in names:

        ind = [i for i, k in enumerate(varis) if k.name == v]
        assert len(ind) == 1, f'Each input variable must appear exactly once in M.variables: {v} appears {len(idx)} times.'
        idx.append(ind[0])

    return idx


def get_prod_idx(cpms, varis):
    """
    Returns index for product operation

    Parameters
    ----------
    cpms:
    varis:
    """

    assert isinstance(cpms, (list,  dict)), 'cpms should be a list or dict'
    if isinstance(cpms, dict):
        cpms = cpms.values()

    assert isinstance(varis, (list,  dict)), 'varis should be a list or dict'
    if isinstance(varis, dict):
        varis = varis.values()

    idx = []
    for x in cpms:
        val = set(x.variables[x.no_child:]).difference(varis)
        idx.append(not val)

    try:
        # take integer from the list
        return np.where(idx)[0][0]

    except IndexError as e:
        print(f'CPMs include undefined parent node: {idx}')


def get_sample_order(cpms):

    if isinstance(cpms, dict):

        cpms = list(cpms.values())

    ncpms = len(cpms)
    cpms_ = copy.deepcopy(cpms)
    cpms_idx = list(range(ncpms))

    sample_order = []
    sample_vars = []
    var_add_order = []

    for i in range(ncpms):

        cpm_prod_idx = get_prod_idx(cpms_, sample_vars)

        if cpm_prod_idx is not None:
            sample_order.append(cpms_idx[cpm_prod_idx])
            cpm_prod = cpms_[cpm_prod_idx]

            vars_prod = cpm_prod.variables[:cpm_prod.no_child]

            if set(sample_vars).intersection(vars_prod):
                print('Given Cpms must not have common child nodes')
            else:
                [sample_vars.append(x) for x in vars_prod]

            try:
                var_add_order = np.append(
                    var_add_order,
                    i*np.ones(len(vars_prod)))
            except NameError:
                var_add_order = i*np.ones(len(vars_prod))

            cpms_.pop(cpm_prod_idx)
            cpms_idx.pop(cpm_prod_idx)

    return sample_order, sample_vars, var_add_order


def mcs_product(cpms, nsample, is_scalar=True):
    """
    Returns an instance of Cpm by MC based product operation

    Parameters
    ----------
    cpms: a list of instances of Cpm
    nsample: number of samples
    is_scalar: True if a prob is given as a scalar (all multiplied into one number); False if given as a list for each sampled variables
    """
    sample_order, sample_vars, var_add_order = get_sample_order(cpms)

    nvars = len(sample_vars)
    C_prod = np.zeros((nsample, nvars), dtype=int)

    if is_scalar:
        q_prod = np.zeros((nsample, 1))
    else:
        q_prod = np.zeros((nsample, nvars))

    sample_idx_prod = np.arange(nsample)

    for i in sample_idx_prod:

        sample, sample_prob = single_sample(cpms, sample_order, sample_vars, var_add_order, [i], is_scalar)
        C_prod[i,:] = sample
        if is_scalar:
            q_prod[i] = sample_prob
        else:
            q_prod[i,:] = sample_prob

    if is_scalar:
        q_prod = q_prod[:,::-1]

    return cpm.Cpm(variables=sample_vars[::-1],
               no_child=nvars,
               C = [],
               p = [],
               Cs=C_prod[:, ::-1],
               q=q_prod,
               sample_idx=sample_idx_prod)


def single_sample(cpms, sample_order, sample_vars, var_add_order, sample_idx, is_scalar=True):
    """
    sample from cpms

    parameters:
        cpms: list or dict
        sample_order: list-like
        sample_vars: list-like
        var_add_order: list-like
        sample_idx: list
        is_scalar: True if a prob is given as a scalar (all multiplied into one number); False if given as a list for each sampled variables
    """
    assert isinstance(sample_vars, list), 'should be a list'

    if isinstance(var_add_order, list):
        var_add_order = np.array(var_add_order)

    if isinstance(cpms, dict):
        cpms = cpms.values()

    sample = np.zeros(len(sample_vars), dtype=int)

    for i, (j, M) in enumerate(zip(sample_order, cpms)):

        #FIXME
        cnd_vars = [x for x, y in zip(sample_vars, sample) if y > 0]
        cnd_states = sample[sample > 0]

        M = M.condition(cnd_vars=cnd_vars,
                        cnd_states=cnd_states)

        #if (sample_idx == [1]) and any(M.p.sum(axis=0) != 1):
        #    print('Given probability vector does not sum to 1')

        weight = M.p.flatten()/M.p.sum(axis=0)
        try:
            irow = np.random.choice(range(len(M.p)), size=1, p=weight)[0]
        except ValueError:
            print('ValueError')
        else:
            if is_scalar:
                try:
                    sample_prob += np.log(M.p[[irow]])
                except NameError:
                    sample_prob = np.log(M.p[[irow]])
            else:
                try:
                    sample_prob = np.append(
                        sample_prob,
                        np.log(M.p[[irow]]))
                except NameError:
                    sample_prob = np.array(np.log(M.p[[irow]]))

            idx = M.C[irow, :M.no_child]
            sample[var_add_order == i] = idx

    try:
        sample_prob = np.exp(sample_prob)
    except NameError:
        return sample, None
    else:
        return sample, sample_prob


def rejection_sampling_sys(cpms, sys_name, sys_fun, nsamp_cov, sys_st_monitor = None, known_prob=0.0, sys_st_prob = 0.0, rand_seed = None):
    """
    Perform rejection sampling on cpms w.r.t. given C
    INPUT:
    - cpms: a list/dictionary of cpms (including system event)
    - sys_name: the key in cpms that represents a system event (a rejection sampling is performed to avoid Csys)
    - nsamp_cov: either number of samples (if integer) or target c.o.v. (if float value)
    - sys_fun: a function (input: comp_st as dictionary and output: sys_val, sys_st)
    - isRejected: True if instances in C be rejected; False if instances not in C be.
    - sys_st_monitor: an integer representing system state to be reference to compute c.o.v. (only necessary when nsamp_cov indicates c.o.v.)
    - known_prob: a float representing already known probability (i.e. those represented by cpms[sys_name].C)
    - sys_st_prob: a float representing already known probility of sys_st
    - rand_seed: a scalar representing random ssed

    OUTPUT:
    - cpms: a list;dictionary of cpms with samples added
    - result: a dictionary including the summary of sampling result
    """
    if rand_seed:
        random.seed(rand_seed)

    assert isinstance(nsamp_cov, (float, int)), 'nsamp_cov must be either an integer (considered number of samples) or a float value (target c.o.v.)'
    if isinstance(nsamp_cov,int):
        is_nsamp = True
        stop = 0 # stop criterion: number of samples
    elif isinstance(nsamp_cov,float):
        is_nsamp = False
        stop = nsamp_cov + 1 # current c.o.v.

    comp_vars = cpms[sys_name].variables[cpms[sys_name].no_child:]
    comp_names = [x.name for x in comp_vars]
    C_reject = cpms[sys_name].C[:,cpms[sys_name].no_child:]

    cpms_no_sys = {}
    for k, m in cpms.items():
        if ( k != sys_name ) and ( set([v.name for v in m.variables[:m.no_child]]) & set([v.name for v in cpms[sys_name].variables]) ):
            cpms_no_sys[k] = m

    sample_order, sample_vars, var_add_order = get_sample_order(cpms_no_sys)
    sample_vars_str = [x.name for x in sample_vars]
    try:
        comp_vars_loc = [sample_vars_str.index(x) for x in comp_names]
    except ValueError:
        comp_vars_loc = []
    else:
        cpms_v_idxs_ = {k: get_var_ind(sample_vars, [y.name for y in x.variables[:x.no_child]])
            for k, x in cpms_no_sys.items()}
        cpms_v_idxs = {k: [cpms_v_idxs_[y.name][0]
            for y in x.variables] for k, x in cpms_no_sys.items()}

    nsamp_tot = 0 # total number of samples (including rejected ones)
    nsamp = 0 # accepted samples
    nfail = 0 # number of occurrences in system state to be monitored
    sys_vals = []
    samples = np.empty((0, len(sample_vars)), dtype=int)
    samples_sys = np.empty((0, 1), dtype=int)
    sample_probs = np.empty((0, len(sample_vars)), dtype=float)

    pf, cov = 0.0, 1.0
    while (is_nsamp and stop < nsamp_cov) or (not is_nsamp and stop > nsamp_cov):

        nsamp_tot += 1

        sample, sample_prob = single_sample(cpms_no_sys, sample_order, sample_vars, var_add_order, [nsamp], is_scalar=False)

        sample_comp = sample[comp_vars_loc]
        is_cpt = cpm.iscompatible(C_reject, comp_vars, comp_names, sample_comp)

        if (~is_cpt).all():

            comp_st = {x:sample_comp[i] for i,x in enumerate(comp_names)}
            sys_val, sys_st = sys_fun(comp_st)

            samples = np.vstack((samples, sample))
            samples_sys = np.vstack((samples_sys, [sys_st]))
            sample_probs = np.vstack((sample_probs, sample_prob))

            nsamp += 1
            if is_nsamp:
                stop += 1
            else:
                if sys_val == sys_st_monitor:
                    nfail +=1

                if nsamp > 9:
                    prior = 0.01
                    a, b = prior + nfail, prior + (nsamp-nfail) # Bayesian estimation assuming beta conjucate distribution
                    pf_s = a / (a+b)
                    var_s = a*b / (a+b)**2 / (a+b+1)
                    std_s = np.sqrt(var_s)

                    pf = sys_st_prob + (1 - known_prob) *pf_s
                    std = (1 - known_prob) * std_s

                    cov = std/pf
                    stop = cov
                else:

                    stop = nsamp_cov + 1 # do not stop the sampling until a few samples are secured

            sys_vals.append(sys_val)

        if nsamp_tot % 1000 == 0: # For monitoring
            print(f'[No. of samples] (total, accept): {nsamp_tot, nsamp}')
            print(f'(pf, c.o.v.): {pf, cov}')


    #Result
    ## Allocate samples to CPMs
    for k, M in cpms_no_sys.items():

        col_loc = cpms_v_idxs[k]
        col_loc_c = cpms_v_idxs[k][:M.no_child]

        Cs = samples[:,col_loc]
        q = sample_probs[:,col_loc_c]

        M2 = cpm.Cpm(variables=M.variables,
                 no_child = M.no_child,
                 C = M.C,
                 p = M.p,
                 Cs = Cs,
                 q = q,
                 sample_idx = np.arange(nsamp))

        cpms[k] = M2

    Cs_sys = np.hstack((samples_sys,samples[:,comp_vars_loc]))
    M = cpms[sys_name]
    M2 = cpm.Cpm(variables=M.variables,
             no_child = M.no_child,
             C = M.C,
             p = M.p,
             Cs = Cs_sys,
             q = np.ones((nsamp,1)), # assuming a deterministic system function
             sample_idx = np.arange(nsamp))

    cpms[sys_name] = M2

    result = {'pf': pf,
              'cov': cov,
              'nsamp_tot': nsamp_tot,
              'nsamp': nsamp,
              'sys_vals': sys_vals}

    return cpms, result


def isinscope(idx, cpms):
    """
    return list of boolean
    idx: list of index
    cpms: list or dict of CPMs
    """
    assert isinstance(idx, list), 'idx should be a list'

    if isinstance(cpms, dict):
        cpms = cpms.values()

    isin = np.zeros((len(cpms), 1), dtype=bool)

    for i in idx:

        if isinstance(i, str):
            flag = np.array([cpm.ismember([i], [x.name for x in M.variables])[0] for M in cpms])
        else:
            flag = np.array([cpm.ismember([i], M.variables)[0] for M in cpms])
        isin = isin | flag

    return isin


def variable_elim(cpms, var_elim, prod=True):
    """
    cpms: list or dict of cpms

    var_elim_order:
    """
    if isinstance(cpms, dict):
        cpms = list(cpms.values())
    else:
        cpms = copy.deepcopy(cpms)

    assert isinstance(var_elim, list), 'var_elim should be a list of variables'

    for _var in var_elim:

        isin = isinscope([_var], cpms)

        sel = [y for x, y in zip(isin, cpms) if x]
        mult = cpm.product(sel)
        mult = mult.sum([_var])

        cpms = [y for x, y in zip(isin, cpms) if x == False]
        cpms.insert(0, mult)

    if prod:
        cpms = cpm.product(cpms)

    return cpms


def variable_elim_cond(cpms, varis, cpms_cond):
    """
    [INPUT]
    cpms: list or dict of cpms for VE
    varis: a list of variables "names" to be eliminated by VE
    cpms_cond: list or dict of cpms for conditioning

    [OUTPUT]
    M: a cpm with varis and variables of cpms_cond eliminated.
    """
    if isinstance(cpms, dict):
        cpms = list(cpms.values())
    else:
        cpms = copy.deepcopy(cpms)

    if isinstance(cpms_cond, dict):
        cpms_cond = list(cpms_cond.values())
    else:
        cpms_cond = copy.deepcopy(cpms_cond)

    assert isinstance(varis, list), 'varis should be a list of variable names'

    c_names = list(set([v.name for x in cpms_cond for v in x.variables]))

    cpms_prod = cpm.product(cpms_cond)

    for cx in cpms_prod.C:

        _cpms_prod = cpms_prod.condition(cpms_prod.variables, cx)
        _cpms_cond = condition(cpms, cpms_prod.variables, cx)

        _cpm = variable_elim(_cpms_cond, varis).product(_cpms_prod)

        idx = _cpm.get_col_ind(c_names)

        try:
            idv
        except NameError:
            idv = [i for i in range(len(_cpm.variables)) if i not in idx]
        finally:
            if idv:
                cpm2 = cpm.Cpm([_cpm.variables[v] for v in idv], _cpm.no_child-len(idx), _cpm.C[:, idv], _cpm.p)

                try:
                    M = M.merge(cpm2)
                except NameError:
                    M = copy.deepcopy(cpm2)
            else:
                del idv

    try:
        M
    except NameError:
        M = None
    finally:
        return M


def get_variables(cpms, variables):

    assert isinstance(cpms, (dict, list)), f'M should be either dict or list but {type(M)}'
    assert isinstance(variables, list), f'variables should be a list but {type(variables)}'
    assert all(isinstance(x, str) for x in variables), f'variables should be a list of str but {type(variables[0])}'

    res = []
    remain = variables[:]

    if isinstance(cpms, dict):
        cpms = cpms.values()

    for M in cpms:
        names = [x.name for x in M.variables]
        i = 0
        while i < len(remain):
            if remain[i] in names:
                res.append(M.get_variables(remain[i]))
                remain.remove(remain[i])
                #i -= 1
            else:
                i += 1

    assert len(res) == len(variables), f'not all variables found in M: {set(variables).difference([x.name for x in res])}'

    return sorted(res, key=lambda x: variables.index(x.name))

"""#FIXME: UNUSED
def append(cpm1, cpm2):
    '''
    return a list of combined cpm1 and cpm2
    cpm1 should be a list or dict
    cpm2 should be a list or dict
    '''

    assert isinstance(cpm1, (list, dict)), 'cpm1 should be a list or dict'
    assert isinstance(cpm2, (list, dict)), 'cpm2 should be a list or dict'

    if isinstance(cpm1, dict):
        cpm1 = list(cpm1.values())

    if isinstance(cpm2, dict):
        cpm2 = list(cpm2.values())

    assert len(cpm1) == len(cpm2), 'Given CPMs have different lengths'
"""


def prod_Msys_and_Mcomps(Msys, Mcomps_list):
    """


    """

    # New CPM's scope
    assert all(x.no_child == 1 for x in Mcomps_list), 'All CPMs of component events must have one child node.'

    sys_vars_ch = Msys.variables[:Msys.no_child] # child variables
    sys_vars_par = [] # parent variables

    comp_names = [x.variables[0].name for x in Mcomps_list]
    mult_idxs = []
    for v in Msys.variables[Msys.no_child:]:

        if v.name in comp_names:
            idx = comp_names.index(v.name)
            sys_vars_ch += [v]
            mult_idxs += [idx]
        else:
            sys_vars_par += [v]

    cond_vars, cond_st = [], []
    if len(Mcomps_list[0].variables) > Mcomps_list[0].no_child: # component CPMs are conditional distributions.
        cond_vars = Mcomps_list[0].variables[Mcomps_list[0].no_child:]

        assert all(x.variables[1:] == cond_vars for x in Mcomps_list), "All CPMS of component events must have the same conditional variables."

        cond_st = Mcomps_list[0].C[0][Mcomps_list[0].no_child:]
        for m in Mcomps_list:
            assert all( c[1:]==cond_st for c in m.C ), "All instances of component events must be conditioned on the same states."


    if Msys.C.size:
        C = np.empty_like(Msys.C)

        for i, v in enumerate(sys_vars_ch):
            col_idx = Msys.get_col_ind([v.name])
            C[:,i] = np.squeeze( Msys.C[:,col_idx] )

        no_ch = len(sys_vars_ch)
        for i, v in enumerate(sys_vars_par):
            col_idx = Msys.get_col_ind([v.name])
            C[:,i+no_ch] = np.squeeze( Msys.C[:,col_idx] )

        if len(cond_vars) > 0:
            for v, s in zip(cond_vars, cond_st):
                if v not in sys_vars_par:
                    sys_vars_par += [v]

                    cond_tiles = np.tile([s], (C.shape[0], 1))
                    C = np.concatenate((C, cond_tiles), axis=1)

        p = Msys.p.copy()
        for i, idx in enumerate(mult_idxs):
            m_i = Mcomps_list[idx]

            c1 = [c[0] for c in m_i.C]
            for j in range(len(p)):
                st = C[j][Msys.no_child + i]
                p_st = 0.0
                for k in m_i.variables[0].B[st]:
                    p_st += m_i.p[c1.index(k)][0]

                p[j] = p[j] * p_st

    if Msys.Cs.size and all(M.Cs.size for M in Mcomps_list):
        cpms_noC = {}
        cpms_noC[Msys.variables[0]] = copy.deepcopy( Msys )
        for v in Msys.variables[Msys.no_child:]:
            if v not in cond_vars:
                m_x = next((m for m in Mcomps_list if m.variables[0].name == v.name), None)
                assert m_x is not None, f'There is no cpm found for component event {v}'
                cpms_noC[v.name] = copy.deepcopy(m_x)

        for k, v in cpms_noC.items():
            v.C, v.p = np.empty((0,len(v.variables))), np.empty((0,1))

        cpm_sys2 = cpm.product(cpms_noC)
        Cs = cpm_sys2.Cs
        q = cpm_sys2.q
        ps = cpm_sys2.ps
        sample_idx = cpm_sys2.sample_idx

        v_idx = cpm_sys2.get_col_ind([v.name for v in sys_vars_ch+sys_vars_par])
        Cs = Cs[:, v_idx]

    else:
        Cs = np.empty((0, len(Msys.variables)))
        q, ps, sample_idx = np.empty((0,1)), np.empty((0,1)), np.empty((0,1))

    return cpm.Cpm(sys_vars_ch+sys_vars_par, len(sys_vars_ch), C, p, Cs, q, ps, sample_idx)


def get_inf_vars(cpms, varis, ve_ord=None):

    """
    INPUT:
    - cpms: a list of CPMs
    - varis: a list of variable names, whose marginal distributions are of interest
    - ve_ord (optional): a list of variable names, representing a VE order. The output list of vars_inf is sorted accordingly.
    OUPUT:
    - varis_inf: a list of variable names
    """

    def get_ord_inf(x, alist):
        try:
            return alist.index(x)
        except ValueError:
            return len(alist)

    if isinstance(varis, str):
        varis = [varis]

    assert isinstance(varis, list), f'varis must be a list: {type(varis)}'

    varis_inf = [] # relevant variables for inference
    varis_cp = varis[:]

    while varis_cp:

        v = varis_cp[0]
        varis_cp.remove(v)
        varis_inf.append(v)

        scope = [x.name for x in cpms[v].variables[cpms[v].no_child:]] # Scope of v1 (child nodes do not have to be multiplied again)
        for p in scope:
            if p not in varis_inf and p not in varis_cp:
                varis_cp.append(p)

    if ve_ord is not None:
        varis_inf.sort(key=(lambda x: get_ord_inf(x, ve_ord)))

    return varis_inf


def cal_Msys_by_cond_VE(cpms, varis, cond_names, ve_order, sys_name):
    """
    INPUT:
    - cpms: a dictionary of cpms
    - varis: a dictionary of variables
    - cond_names: a list of variables to be conditioned
    - ve_order: a list of variables representing an order of variable elimination
    - sys_name: a system variable's name (NB not list!) **FUTHER RESEARCH REQUIRED: there is no efficient way yet to compute a joint distribution of more than one system event

    OUTPUT:
    - Msys: a cpm containing the marginal distribution of variable 'sys_name'
    """

    vars_inf = get_inf_vars(cpms, [sys_name], ve_order) # inference only ancestors of sys_name
    c_names = [x.name for x in cpms[sys_name].variables[cpms[sys_name].no_child:]]
    ve_names = [x for x in vars_inf if (x in ve_order) and (x not in cond_names) and (x not in c_names)]
    c_jnt_names = [x for x in c_names if x not in ve_order] # if non-empty, a joint distribution (sys, c_jnt) is computed
    c_elm_names = [x for x in c_names if x not in c_jnt_names]

    ve_vars = [varis[v] for v in ve_names if v != sys_name] # other variables

    cpms_inf = {v: cpms[v] for v in ve_names + c_names + c_jnt_names}
    cpms_inf[sys_name] = cpms[sys_name]
    cond_cpms = [cpms[v] for v in cond_names]

    M_cond = cpm.product(cond_cpms)
    n_crows = len(M_cond.C)

    for i in range(n_crows):
        #m1 = M_cond.get_subset([i])
        m1 = condition(M_cond, M_cond.variables, M_cond.C[i,:])
        m1 = m1[0]
        VE_cpms_m1 = condition(cpms_inf, m1.variables, m1.C[0])

        for x_name in c_jnt_names:
            VE_cpms_m1[sys_name] = VE_cpms_m1[sys_name].product(VE_cpms_m1[x_name])
            del VE_cpms_m1[x_name]

        VE_cpms_m1_no_sys = {k: v for k,v in VE_cpms_m1.items() if k != sys_name}
        cpms_comps = variable_elim(VE_cpms_m1_no_sys, ve_vars, prod=False)

        m_m1 = prod_Msys_and_Mcomps(VE_cpms_m1[sys_name], cpms_comps)
        m_m1 = m_m1.sum(c_elm_names)

        m_m1 = m_m1.product(m1)
        m_m1 = m_m1.sum(cond_names)

        if i < 1:
            Msys = copy.deepcopy(m_m1)
        else:
            Msys = Msys.merge(m_m1)

    return Msys


# quantify cpms for standard system types
def sys_max_val(name, vars_p, B_flag='store'):

    # Reference: variant of Algorithm MBN-quant-MSSP in Byun, J. E., & Song, J. (2021). Generalized matrix-based Bayesian network for multi-state systems. Reliability Engineering & System Safety, 211, 107468.
    # C_max is quantified for a system whose value is determined as the maximum values of variables in vars_p
    """
    INPUT:
    vars_p: a list of variables (the parent nodes of the node of interest)
    OUTPUT:
    cpm_new: a new CPM representing the system node
    var_new: a new variable representing the system node
    """

    def get_mv(var): # get minimum values
        return min(var.values)

    vars_p_s = copy.deepcopy(vars_p)
    vars_p_s.sort(key=get_mv) # sort variables to minimise # of C's rows

    C_new = np.empty(shape=(0, 1 + len(vars_p)), dtype='int32') # var: [X] + vars_p

    vals_new = []
    for i, p in enumerate(vars_p_s):

        vs_p = copy.deepcopy(vars_p_s[i].values)
        vs_p.sort()

        for v in vs_p:
            c_i = np.zeros(shape=(1, 1 + len(vars_p)), dtype='int32')

            j = vars_p.index(p)
            c_i[0][j + 1] = p.values.index(v)

            add = True
            for i2, p2 in enumerate(vars_p_s):
                if i != i2:
                    if i2 < i:
                        vs_i2 = {y for y, z in enumerate(p2.values) if z < v}

                    if i2 > i:
                        vs_i2 = {y for y, z in enumerate(p2.values) if z <= v}

                    if len(vs_i2) < 1:
                        add = False
                        break
                    else:
                        st_i2 = p2.B.index(vs_i2)

                        j2 = vars_p.index(p2)
                        c_i[0][j2 + 1] = st_i2

            if add:
                if v not in vals_new:
                    vals_new.append(v)
                c_i[0][0] = vals_new.index(v)
                C_new = np.vstack([C_new, c_i])

    vals_new.sort()

    var_new = variable.Variable(name=name, values=vals_new, B_flag='store')
    cpm_new = cpm.Cpm(variables=[var_new] + vars_p, no_child = 1,
                      C=C_new, p=np.ones(shape=(len(C_new), 1), dtype='float64'))

    return cpm_new, var_new


def sys_min_val(name, vars_p, B_flag='store'):

    # Reference: variant of Algorithm MBN-quant-MSSP in Byun, J. E., & Song, J. (2021). Generalized matrix-based Bayesian network for multi-state systems. Reliability Engineering & System Safety, 211, 107468.
    # C_max is quantified for a system whose value is determined as the minimum values of variables in vars_p
    """
    INPUT:
    vars_p: a list of variables (the parent nodes of the node of interest)
    OUTPUT:
    cpm_new: a new CPM representing the system node
    var_new: a new variable representing the system node
    """

    def get_mv(var): # get maximum values
        return max(var.values)

    vars_p_s = copy.deepcopy(vars_p)
    vars_p_s.sort(key=get_mv) # sort variables to minimise # of C's rows

    C_new = np.empty(shape=(0, 1 + len(vars_p)), dtype='int32') # var: [X] + vars_p

    vals_new = []
    for i, p in enumerate(vars_p_s):

        vs_p = copy.deepcopy(vars_p_s[i].values)
        vs_p.sort()

        for v in vs_p:
            c_i = np.zeros(shape=(1, 1 + len(vars_p)), dtype='int32')

            j = vars_p.index(p)
            c_i[0][j + 1] = p.values.index(v)

            add = True
            for i2, p2 in enumerate(vars_p_s):
                if i != i2:
                    if i2 < i:
                        vs_i2 = {y for y, z in enumerate(p2.values) if z > v}

                    if i2 > i:
                        vs_i2 = {y for y, z in enumerate(p2.values) if z >= v}


                    if len(vs_i2) < 1:
                        add = False
                        break
                    else:
                        st_i2 = p2.B.index(vs_i2)

                        j2 = vars_p.index(p2)
                        c_i[0][j2 + 1] = st_i2

            if add:
                if v not in vals_new:
                    vals_new.append(v)

                c_i[0][0] = vals_new.index(v)
                C_new = np.vstack([C_new, c_i])

    vals_new.sort()

    var_new = variable.Variable(name=name, values=vals_new, B_flag='store')
    cpm_new = cpm.Cpm(variables=[var_new] + vars_p, no_child = 1,
                      C=C_new, p=np.ones(shape=(len(C_new), 1), dtype='float64') )

    return cpm_new, var_new


def max_flow(comps_st, target_flow, od_pair, edges, varis): # maximum flow analysis

    """
    INPUT:
    comps_st: a dictionary of component states
    target_flow: an integer
    od_pair: a tuple
    edges: a dictionary of edge connectivities
    varis: a dictionary of mbnpy.variable

    OUTPUT:
    f_val: any quantity that indicates a level of system performance (for user information's sake)
    sys_st: either 's' or 'f'
    min_comps_st: a dictionary of minimal survival rules in case 's', None in case of 'f'

    Example:
    nodes = {'n1': (0, 0),
         'n2': (1, 1),
         'n3': (1, -1),
         'n4': (2, 0)}

    edges = {'e1': ['n1', 'n2'],
            'e2': ['n1', 'n3'],
            'e3': ['n2', 'n3'],
            'e4': ['n2', 'n4'],
            'e5': ['n3', 'n4']}

    od_pair=('n1','n4')

    varis = {}
    for k, v in edges.items():
        varis[k] = variable.Variable( name=k, values = [0, 1, 2]) # values: edge flow capacity
    """

    G = nx.Graph()
    for k,x in comps_st.items():
        G.add_edge(edges[k][0], edges[k][1], capacity=varis[k].values[x])
    G.add_edge(od_pair[1], 'new_t', capacity=target_flow)

    f_val, f_dict = nx.maximum_flow(G,od_pair[0], 'new_t', capacity='capacity', flow_func=shortest_augmenting_path)

    if f_val >= target_flow:
        sys_st = 's'

        min_comps_st = {}
        for k, x in comps_st.items():
            k_flow = max([f_dict[edges[k][0]][edges[k][1]], f_dict[edges[k][1]][edges[k][0]]])
            if k_flow > 0: # the edge plays a role in an associated survival rule
                index = next((i for i,x in enumerate(varis[k].values) if x >= k_flow), None)
                min_comps_st[k] = index

    else:
        sys_st = 'f'
        min_comps_st = None

    return f_val, sys_st, min_comps_st




from itertools import groupby
import logging
#import configparser
import pandas as pd
import numpy as np
from scipy import stats

#from pathlib import Path


def all_equal(iterable):
    "Returns True if all the elements are equal to each other"
    g = groupby(iterable)
    return next(g, True) and not next(g, False)

# read node data
def read_nodes(file_node):

    df = pd.read_csv(file_node, index_col=0).to_dict('index')

    return {k: (v['x'], v['y']) for k, v in df.items()}


def get_rat_dist(k1, k2, q, a1, a2, b1, b2, p):
    """
    Computes distribution of a ratio of (k1 + q1*X1) / (k1 + k2 + q*(X1+X2)),
    where X1~Beta(a1, b1) and X2~beta(a2,b2)

    Output:
    - mean: a float; mean value
    - std: a float; standard deviation
    - interval: a tuple; [0.5p, 1-0.5p] confidence interval

    Personal note: Given up deriving analytical form of the distribution. Future research?
    """
    # Check if all input values are positive
    if not all(i >= 0 for i in [k1, k2, a1, a2, b1, b2]):
        raise ValueError("All input values must be non-negative")

    if not (p > 0 and p < 1):
        raise ValueError("p (confidence interval prob.) must be between 0 and 1.")

    n_samp = int(1e6) # many samples
    # Generate samples
    x1 = np.random.beta(a=a1, b=b1, size=n_samp)
    x2 = np.random.beta(a=a2, b=b2, size=n_samp)

    # Calculate the mean difference and standard deviation of the difference
    ratio = (q*x1 + k1) / (q*(x1 + x2) + k1 + k2)
    mean_rat = np.mean(ratio)
    std_rat = np.std(ratio, ddof=1)

    # Calculate the 95% confidence interval for the difference
    ci_lower = np.percentile(ratio, 50*(1 - p))
    ci_upper = np.percentile(ratio, 100-50*(1 - p))

    return mean_rat, std_rat, (ci_lower, ci_upper)




import numpy as np

from mbnpy import bnb_fns


def test_bnb_sys1():

    comp_states = [1, 1, 1, 1, 1, 1]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2}
    #pdb.set_trace()
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state== 3-1
    assert val== np.inf
    assert result== {'path': []}


def test_bnb_sys2():

    comp_states = [1, 2, 1, 1, 1, 1]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2}
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state== 1-1
    assert val== 0.0901
    assert result=={'path': [2]}


def test_bnb_sys3():

    comp_states = [2, 2, 2, 2, 2, 2]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2}
    #pdb.set_trace()
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state==1-1
    assert val==0.0901
    assert result=={'path': [2]}


def test_bnb_sys4():

    comp_states = [1, 2, 2, 2, 2, 2]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2
            }
    #pdb.set_trace()
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state==1-1
    assert val==0.0901
    assert result=={'path': [2]}


def test_bnb_sys5():

    comp_states = [1, 1, 2, 2, 2, 2]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2}
    #pdb.set_trace()
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state== 3-1
    assert val== np.inf
    assert result== {'path': []}


def test_bnb_sys6():
    # 2: survival, 1: failure
    comp_states = [2, 1, 2, 1, 1, 1]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            'max_state': 2,
            }
    #pdb.set_trace()
    state, val, result = bnb_fns.bnb_sys(comp_states, info)

    assert state==2-1
    assert val==0.2401
    assert result=={'path': [3, 1]}


def test_bnb_next_comp1():

    cand_comps = [1, 2, 3, 4, 5, 6]
    down_res = []
    up_res = [3, 1]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': [1, 2, 3, 4, 5, 6],
            }

    next_comp = bnb_fns.bnb_next_comp(cand_comps, down_res, up_res, info)

    assert next_comp== 1


def test_bnb_next_comp2():

    cand_comps = [2, 3, 4, 5, 6]
    down_res = []
    up_res = [2]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': np.array([1, 2, 3, 4, 5, 6]),
            }

    next_comp = bnb_fns.bnb_next_comp(cand_comps, down_res, up_res, info)

    assert next_comp== 2


def test_bnb_next_comp3():

    cand_comps = [3, 4, 5, 6]
    down_res = []
    up_res = [3, 1]
    info = {'path': [[2], [3, 1]],
            'time': np.array([0.0901, 0.2401]),
            'arcs': np.array([1, 2, 3, 4, 5, 6]),
            }

    next_comp = bnb_fns.bnb_next_comp(cand_comps, down_res, up_res, info)

    assert next_comp== 3





from .dag import DAG

class Pipeline:
    """Create a pipeline by chaining multiple tasks and identifying dependencies."""

    def __init__(self):
        self.tasks = DAG()


    def task(self, depends_on=None):
        """Add new task to the pipeline and specify dependency task (optional)."""
        def inner(func):
            if depends_on:
                self.tasks.add(depends_on, func)
            else:
                self.tasks.add(func)
            return func
        return inner


    def run(self, *args):
        """Execute the pipeline and return each task results."""
        sorted_tasks = self.tasks.sort()
        completed = {}

        for task in sorted_tasks:
            for depend_on, nodes in self.tasks.graph.items():
                if task in nodes:
                    completed[task] = task(completed[depend_on])
            if task not in completed:
                if sorted_tasks.index(task) == 0:
                    completed[task] = task(*args)
                else:
                    completed[task] = task()

        return completed



# coding: utf-8

import pytest
import copy
import time
import pdb
import pickle
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from pathlib import Path
from collections import Counter

from mbnpy import trans, branch, variable, cpm, brc, inference


HOME = Path(__file__).parent


def compare_list_of_sets(a, b):

    return set([tuple(x) for x in a]) == set([tuple(x) for x in b])


@pytest.fixture(scope='package')
def setup_inference(main_sys, setup_brs):

    G, _, arcs, _, _ = main_sys

    d_varis, brs = setup_brs
    varis = copy.deepcopy(d_varis)

    # Component events
    cpms = {}
    for k, v in arcs.items():
        cpms[k] = cpm.Cpm(variables=[varis[k]],
                          no_child = 1,
                          C = np.array([0, 1, 2]),
                          p = [0.1, 0.2, 0.7])

    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}
    csys, varis = brc.get_csys(brs, varis, st_br_to_cs)

    # Damage observation
    C_o = np.array([[0, 0], [1, 0], [2, 0],
                    [0, 1], [1, 1], [2, 1],
                    [0, 2], [1, 2], [2, 2]])

    p_o = np.array([0.95, 0.04, 0.01,
                    0.3, 0.5, 0.2,
                    0.01, 0.19, 0.8]).T

    for i, k in enumerate(arcs, 1):
        name = f'o{i}'
        varis[name] = variable.Variable(name=name, values = [0, 1, 2])
        cpms[name] = cpm.Cpm(variables=[varis[name], varis[k]], no_child=1, C=C_o, p=p_o)

    # add observations
    added = np.ones(shape=(csys.shape[0], len(arcs)), dtype=np.int8) * 6
    csys = np.append(csys, added, axis=1)

    # add sys
    varis['sys'] = variable.Variable('sys', ['f', 's', 'u'])
    cpm_sys_vname = [f'e{i}' for i in range(1, len(arcs) + 1)] + [f'o{i}' for i in range(1, len(arcs) + 1)] # observations first, components later
    cpm_sys_vname.insert(0, 'sys')
    cpms['sys'] = cpm.Cpm(
        variables=[varis[k] for k in cpm_sys_vname],
        no_child = 1,
        C = csys,
        p = np.ones((len(csys), 1), int))

    var_elim_order_name = [f'o{i}' for i in range(1, len(arcs) + 1)] + [f'e{i}' for i in range(1, len(arcs) + 1)] # observations first, components later
    var_elim_order = [varis[k] for k in var_elim_order_name]

    return cpms, varis, var_elim_order, arcs


@pytest.fixture(scope='package')
def main_sys(data_bridge):

    node_coords = data_bridge['node_coords']
    arcs = data_bridge['arcs']
    arcs_avg_kmh = data_bridge['arcs_avg_kmh']

    # od_pair is different from setup_bridge in test_trans.py
    od_pair = ('n1', 'n3')

    arc_lens_km = trans.get_arcs_length(arcs, node_coords)
    arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    G = nx.Graph()
    for k, v in arcs.items():
        G.add_edge(v[0], v[1], label=k, key=k, weight=arc_times_h[k])

    [G.add_node(f'n{k}') for k in range(1, 6)]

    # add weight
    for n1, n2, e in G.edges(data=True):
        G[n1][n2]['weight'] = arc_times_h[e['label']]

    # Component events
    #no_arc_st = 3 # number of component states 
    varis = {}
    delay_rat = [10, 2, 1] # delay in travel time given each component state (ratio)
    for k, v in arcs.items():
        varis[k] = variable.Variable(name=k, values = [arc_times_h[k]*np.float64(x) for x in delay_rat])
    probs = {k: v.values for k, v in varis.items()}

    return G, od_pair, arcs, varis, probs


@pytest.fixture(scope='package')
def setup_brs(main_sys):

    G, od_pair, arcs, d_varis, probs = main_sys

    varis = copy.deepcopy(d_varis)

    # Intact state of component vector: zero-based index 
    comps_st_itc = {k: len(v.values) - 1 for k, v in varis.items()} # intact state (i.e. the highest state)
    d_time_itc, path_n, path_itc = trans.get_time_and_path_given_comps(comps_st_itc, G, od_pair, varis)

    # defines the system failure event
    thres = 2
    # Given a system function, i.e. sf_min_path, it should be represented by a function that only has "comps_st" as input.
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres * d_time_itc)

    p = [1/3, 1/3, 1/3]
    probs = {k: p for k in varis.keys()}

    file_brs = Path(__file__).parent.joinpath('brs_bridge.pk')
    if file_brs.exists():
        with open(file_brs, 'rb') as fi:
            brs = pickle.load(fi)
            print(f'{file_brs} loaded')
    else:
        # Branch and bound
        output_path = Path(__file__).parent
        #pdb.set_trace()
        brs, rules, sys_res, monitor = brc.run(probs, sys_fun, max_sf=1000, max_nb=100)

    return varis, brs


@pytest.mark.skip('NYI')
@pytest.fixture()
def main_sys_bridge(data_bridge):

    # Network
    _, arcs, arc_times_h = info_bridge

    # od_pair is different from setup_bridge in test_trans.py
    od_pair = ('n5', 'n1')

    #arc_lens_km = trans.get_arcs_length(arcs, node_coords)
    #arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    #comps_name = [k for k in arcs] # *FIXME* this is not necessary if branch's down and up are defined by dictionary (instead of list)

    # Component events
    #no_arc_st = 2 # number of component states 
    #delay_rat = [10, 2, 1] # delay in travel time given each component state (ratio)
    varis = {}
    for k, v in arcs.items():
        #varis[k] = variable.Variable(name=k, B=[{0}, {1}, {2}] , values = [arc_times_h[k]*np.float64(x) for x in delay_rat])
        varis[k] = variable.Variable(name=k, values = [10.0*arc_times_h[k], arc_times_h[k]])

    return od_pair, arcs, varis



@pytest.fixture()
def comps_st_dic():

    comps_st = {}
    expected = {}

    # 0, 1, 2 (higher, better)
    # intact state (i.e. the highest state)
    comps_st[0] = {'e1': 2, 'e2': 2, 'e3': 2, 'e4': 2, 'e5': 2, 'e6': 2}
    expected[0] = (0.1844, 's', {'e2': 2, 'e5': 2}, ['n1', 'n5', 'n3'])

    comps_st[1] = {'e1': 0, 'e2': 1, 'e3': 0, 'e4': 0, 'e5': 0, 'e6': 2}
    expected[1] = (1.12308, 'f', {}, ['n1', 'n5', 'n3'])

    comps_st[2] = {'e1': 2, 'e2': 0, 'e3': 2, 'e4': 2, 'e5': 2, 'e6': 2}
    expected[2] = (0.3344, 's', {'e1': 2, 'e3': 2 , 'e5': 2}, ['n1', 'n2', 'n5', 'n3'])

    comps_st[3] = {'e1': 0, 'e2': 2, 'e3': 1, 'e4': 2, 'e5': 1, 'e6': 1}
    expected[3] = (0.2787, 's', {'e2': 2, 'e5': 1}, ['n1', 'n5', 'n3'])

    comps_st[4] = {'e1': 2, 'e2': 2, 'e3': 0, 'e4': 1, 'e5': 1, 'e6': 2}
    expected[4] = (0.2787, 's', {'e2': 2, 'e5': 1}, ['n1', 'n5', 'n3'])

    comps_st[5] = {'e1':2, 'e2':1, 'e3':2, 'e4':2, 'e5':2, 'e6': 2}
    expected[5] = (0.2746, 's', {'e2': 1, 'e5': 2}, ['n1', 'n5', 'n3'])

    comps_st[6] = {'e1':2, 'e2':2, 'e3':2, 'e4':2, 'e5':1, 'e6': 2}
    expected[6] = (0.26626, 's', {'e2': 2, 'e6': 2, 'e4':2}, ['n1', 'n5', 'n4', 'n3'])

    comps_st[7] = {'e1':1, 'e2':0, 'e3':2, 'e4':2, 'e5':2, 'e6': 2}
    expected[7] = (0.4844, 'f', {}, ['n1', 'n2', 'n5', 'n3'])

    comps_st[7] = {'e1':2, 'e2':0, 'e3':1, 'e4':2, 'e5':2, 'e6': 2}
    expected[7] = (0.4246, 'f', {}, ['n1', 'n2', 'n5', 'n3'])


    return comps_st, expected


def test_get_time_and_path(main_sys, comps_st_dic):

    G, od_pair, arcs, varis, probs = main_sys

    comps_st, expected = comps_st_dic

    for c in comps_st.keys():

        d_time, path, path_e = trans.get_time_and_path_given_comps(comps_st[c], G, od_pair, varis)

        assert pytest.approx(d_time, 0.001) == expected[c][0]
        assert path == expected[c][3]


def test_sf_min_path(main_sys, comps_st_dic):

    G, od_pair, arcs, varis, probs = main_sys
    comps_st, expected = comps_st_dic
    thres = 2*0.1844

    for c in comps_st.keys():

        d_time, path_n, path_e = trans.get_time_and_path_given_comps(comps_st[c], G, od_pair, varis)

        result = trans.sf_min_path(comps_st[c], G, od_pair, varis, thres)

        assert pytest.approx(result[0], 0.001) == expected[c][0]
        assert result[1] == expected[c][1]
        assert result[2] == expected[c][2]


def test_init_branch1():

    rules = {'s':[], 'f': [], 'u': []}
    #worst = {f'e{i}': 0 for i in range(1, 7)}
    #best = {f'e{i}': 2 for i in range(1, 7)}
    varis = {f'e{i}': variable.Variable(name=f'e{i}', values=['1', '2', '3']) for i in range(1, 7)}
    probs = {k: v.values for k, v in varis.items()}
    brs = brc.init_branch(probs, rules)

    assert len(brs) == 1
    assert brs[0].up_state == 'u'
    assert brs[0].down_state == 'u'
    assert brs[0].down == {f'e{i}': 0 for i in range(1, 7)}
    assert brs[0].up == {f'e{i}': 2 for i in range(1, 7)}

    rules = {'s': [{'e2': 2, 'e5': 2}], 'f': [], 'u': []}
    brs = brc.init_branch(probs, rules)

    assert len(brs) == 1
    assert brs[0].up_state == 's'
    assert brs[0].down_state == 'u'
    assert brs[0].down == {f'e{i}': 0 for i in range(1, 7)}
    assert brs[0].up == {f'e{i}': 2 for i in range(1, 7)}

    rules = {'s': [{'e2': 2, 'e5': 2}], 'f': [{f'e{x}': 0 for x in range(1, 7)}], 'u': []}
    brs = brc.init_branch(probs, rules)

    assert len(brs) == 1
    assert brs[0].up_state == 's'
    assert brs[0].down_state == 'f'
    assert brs[0].down == {f'e{i}': 0 for i in range(1, 7)}
    assert brs[0].up == {f'e{i}': 2 for i in range(1, 7)}

    rules = {'s': [{'e2': 2, 'e5': 2}, {'e2': 2, 'e6': 2, 'e4': 2}], 'f': [{f'e{x}': 0 for x in range(1, 7)}], 'u': []}
    brs = brc.init_branch(probs, rules)

    assert len(brs) == 1
    assert brs[0].up_state == 's'
    assert brs[0].down_state == 'f'
    assert brs[0].down == {f'e{i}': 0 for i in range(1, 7)}
    assert brs[0].up == {f'e{i}': 2 for i in range(1, 7)}


def test_get_state0():

    rules = {'s': [], 'f': [], 'u': []}

    cst = {f'e{i}': 2 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 'u'

    cst = {f'e{i}': 0 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 'u'



def test_get_state1():

    rules = {'s': [{'e2': 2, 'e5': 2}], 'f': [], 'u': []}

    cst = {f'e{i}': 2 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 's'

    cst = {f'e{i}': 0 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 'u'


def test_get_state2():

    rules = {'s': [{'e2': 2, 'e5': 2}], 'f': [{'e1': 1, 'e2': 0}], 'u': []}

    cst = {f'e{i}': 2 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 's'

    cst = {f'e{i}': 0 for i in range(1, 7)}
    result = brc.get_state(cst, rules)
    assert result == 'f'


def test_get_state4():

    rules = {'s': [{'e2': 1, 'e5': 2}], 'f': [{'e1': 1, 'e2': 0}], 'u': []}
    cst = {'e1': 0, 'e2': 0, 'e3': 2 , 'e4': 1, 'e5': 0, 'e6': 2}
    result = brc.get_state(cst, rules)
    assert result == 'f'


def test_update_rule_set0():

    rules = {'s': [], 'f': [], 'u': []}
    rule_new = ({'e2':2, 'e5':2}, 's')

    result = brc.update_rule_set(rules, rule_new)
    assert result == {'s': [{'e2': 2, 'e5': 2}], 'f': [], 'u': []}


def test_update_rule_set1():

    rules = {'s': [{'e2': 2, 'e5': 2}], 'f': [], 'u': []}
    rule_new = {f'e{i}':0 for i in range(1, 7)}, 'f'

    result = brc.update_rule_set(rules, rule_new)
    expected = {'s': [{'e2': 2, 'e5': 2}], 'f': [{'e1': 0, 'e2': 0, 'e3': 0, 'e4': 0, 'e5':0, 'e6': 0}], 'u': []}
    assert result == expected


def test_update_rule_set2():

    rules = {'s': [{'e2':2, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}]}
    rule_new = {'e2': 1, 'e5': 2}, 's'

    result = brc.update_rule_set(rules, rule_new)
    expected = {'s': [{'e2':1, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}]}
    assert result == expected

    rule_new = {'e2': 1, 'e5': 0}, 'f'

    result = brc.update_rule_set(rules, rule_new)
    expected = {'s': [{'e2':1, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}]}
    assert result == expected


def test_update_rule_set3():

    rules = {'s': [{'e2':2, 'e5':2}, {'e2':3, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}, {'e2': 0, 'e5': 1}]}
    rule_new = {'e2': 1, 'e5': 2}, 's'
    result = brc.update_rule_set(rules, rule_new)
    expected = {'s': [{'e2':1, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}, {'e2': 0, 'e5': 1}]}
    assert result == expected

    rule_new = {'e2': 1, 'e5': 0}, 'f'

    result = brc.update_rule_set(rules, rule_new)
    expected = {'s': [{'e2':1, 'e5':2}], 'u': [], 'f':[{'e2': 1, 'e5': 1}, {'e2': 0, 'e5': 1}]}
    assert result == expected



def test_run_sys_fn1(main_sys):

    G, od_pair, arcs, varis, probs = main_sys

    thres = 2 * 0.1844
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres)

    cst = {f'e{i}': 2 for i in range(1, 7)}
    rules = []

    rule, sys_res = brc.run_sys_fn(cst, sys_fun, varis)
    np.testing.assert_almost_equal(sys_res['sys_val'].values, np.array([0.18442]), decimal=5)
    assert sys_res['comp_st'].values == [{k: 2 for k in varis.keys()}]
    assert sys_res['comp_st_min'].values == [{'e2': 2, 'e5': 2}]
    assert rule == ({'e2': 2, 'e5': 2}, 's')


def test_run_sys_fn2(main_sys):

    G, od_pair, arcs, varis, probs = main_sys

    thres = 2 * 0.1844
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres)

    cst = {f'e{i}': 0 for i in range(1, 7)}
    rules = [({'e2': 2, 'e5': 2}, 's')]
    rule, sys_res = brc.run_sys_fn(cst, sys_fun, probs)

    np.testing.assert_almost_equal(sys_res['sys_val'].values, np.array([1.8442]), decimal=4)
    assert sys_res['comp_st'].values[0] == {k: 0 for k in varis.keys()}
    assert sys_res['comp_st_min'].values == [{}]
    assert rule == ({k: 0 for k in varis.keys()}, 'f')


def test_run_sys_fn3(main_sys):

    G, od_pair, arcs, varis, probs = main_sys

    thres = 2 * 0.1844
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres)

    cst = {'e1': 2, 'e2': 2, 'e3': 2, 'e4': 2, 'e5': 1, 'e6': 2}
    rules = [({'e2': 2, 'e5': 2}, 's'), ({k: 0 for k in varis.keys()}, 'f')]

    rule, sys_res = brc.run_sys_fn(cst, sys_fun, varis)

    np.testing.assert_almost_equal(sys_res['sys_val'].values, np.array([0.266]), decimal=3)
    assert sys_res['comp_st'].values[0] == {'e1': 2, 'e2': 2, 'e3': 2, 'e4': 2, 'e5': 1, 'e6': 2}
    assert sys_res['comp_st_min'].values == [{'e2': 2, 'e6': 2, 'e4': 2}]
    assert rule == ({'e2': 2, 'e6': 2, 'e4': 2}, 's')


def test_get_composite_state1():

    varis = {}
    varis['e1'] = variable.Variable(name='e1', values=[1.5, 0.3, 0.15])

    states = set([1, 2])
    result = varis['e1'].get_state(states)
    #expected = [{0}, {1}, {2}, {0, 1}, {0, 2}, {1, 2}, {0, 1, 2}]
    #assert compare_list_of_sets(result[0].B, expected)
    assert result == 5


def test_get_composite_state2():

    #od_pair, arcs, varis = main_sys
    varis = {}
    varis['e1'] = variable.Variable(name='e1', values=[1.5, 0.3, 0.15])

    states = set([1, 2])
    result = varis['e1'].get_state(states)

    #expected = [{0}, {1}, {2}, {0, 1}, {0, 2}, {1, 2}, {0, 1, 2}]
    #expected = [{0}, {1}, {2}, {1, 2}]
    #assert compare_list_of_sets(result[0].B, expected)
    assert result == 5


def test_get_composite_state3():

    #od_pair, arcs, varis = main_sys
    varis = {}
    varis['e1'] = variable.Variable(name='e1', values=[1.5, 0.3, 0.15])
    states = set([0, 2])
    result = varis['e1'].get_state(states)

    #expected = [{0}, {1}, {2}, {0, 1}, {0, 2}, {1, 2}, {0, 1, 2}]
    #expected = [{0}, {1}, {2}, {0, 2}]
    #assert compare_list_of_sets(result[0].B, expected)
    assert result == 4


def test_inference1(setup_inference):

    # case 1: no observatio
    cpms, varis, var_elim_order, arcs = setup_inference

    Msys = inference.variable_elim([cpms[v] for v in varis.keys()], var_elim_order )
    np.testing.assert_array_almost_equal(Msys.C, np.array([[0, 1]]).T)
    np.testing.assert_array_almost_equal(Msys.p, np.array([[0.1018, 0.8982]]).T)

    pf_sys = Msys.p[0]
    assert pf_sys == pytest.approx(0.1018, rel=1.0e-3)


def test_inference2(setup_inference):

    # case 2: observation
    cpms, varis, var_elim_order, arcs = setup_inference
    pdb.set_trace()
    cnd_vars = [f'o{i}' for i in range(1, len(arcs) + 1)]
    cnd_states = [1, 1, 0, 1, 0, 1]  # observing e3, e5 failure

    Mobs = inference.condition([cpms[v] for v in varis.keys()], cnd_vars, cnd_states)
    # P(sys, obs)
    Msys_obs = inference.variable_elim(Mobs, var_elim_order)

    np.testing.assert_array_almost_equal(Msys_obs.p[Msys_obs.C==1], np.array([5.515e-5]))
    np.testing.assert_array_almost_equal(Msys_obs.p[Msys_obs.C==0], np.array([2.765e-5]))
    # P(sys=0|obs) = P(sys=0,obs) / P(obs) 
    pf_sys = Msys_obs.p[Msys_obs.C==0] / np.sum(Msys_obs.p)
    assert pf_sys == pytest.approx(0.334, rel=1.0e-3)


@pytest.mark.skip('TODO')
def test_run_brc():
    """
    varis
    probs
    sys_fun
    max_sf
    max_nb

    run_brc(varis, probs, sys_fun, max_sf, max_nb)
    """

    pass

@pytest.mark.skip('TODO')
def test_get_decomp_depth_first():

    max_nb = 1000
    probs = {'e1': 0.1, 'e2': 0.2, 'e3': 0.3}
    varis = {}

    for i in range(1, 4):
        varis[f'e{i}'] = variable.Variable(name=f'e{i}', values=['Fail', 'Survive'])

    varis['od1'] = variable.Variable(name='od1', values=['Fail', 'Survive'])

    rules = {'s': [], 'f': []}

    G = nx.MultiGraph()
    G.add_edge('n1', 'n2', capacity=1)
    G.add_edge('n2', 'n3', 1)
    G.add_edge('n2', 'n3', 1)


def test_get_comp_st1():

    probs = {'e1': {0: 0.1, 1: 0.9},
             'e2': {0: 0.2, 1: 0.8},
             'e3': {0: 0.3, 1: 0.7}}

    varis = {}
    for i in range(1, 4):
        varis[f'e{i}'] = variable.Variable(name=f'e{i}', values=['Fail', 'Survive'])

    brs = [branch.Branch(down={f'e{i}': 0 for i in range(1, 4)},
                         up={f'e{i}': 1 for i in range(1, 4)},
                                down_state = 'u',
                         up_state = 'u',
                         p = 1.0)]
    st = brc.get_comp_st(brs)
    assert st == {'e1': 1, 'e2': 1, 'e3': 1}

    # surv_first = False
    st = brc.get_comp_st(brs, surv_first=False, probs=probs)
    assert st == {'e1': 1, 'e2': 1, 'e3': 1}


def test_get_comp_st2():

    probs = {'e1': {0: 0.1, 1: 0.9},
             'e2': {0: 0.2, 1: 0.8},
             'e3': {0: 0.3, 1: 0.7}}

    brs = [branch.Branch(down={'e1': 1, 'e2': 0, 'e3': 0},
                         up={'e1': 1, 'e2': 0, 'e3': 1},
                         down_state = 'u',
                         up_state = 'u',
                         p = 0.18),
           branch.Branch(down={'e1': 1, 'e2': 1, 'e3': 0},
                         up={'e1': 1, 'e2': 1, 'e3': 1},
                         down_state = 's',
                         up_state = 's',
                         p = 0.72),
           branch.Branch(down={'e1': 0, 'e2': 0, 'e3': 0},
                         up={'e1': 0, 'e2': 1, 'e3': 1},
                         down_state = 'u',
                         up_state = 'u',
                         p = 0.1),
           ]

    st = brc.get_comp_st(brs)
    assert st == {'e1': 1, 'e2': 0, 'e3': 1}


def test_get_connectivity_given_comps4():

    x_star = {'e1': 1, 'e2': 0, 'e3': 1}
    varis = {}
    for i in range(1, 4):
        varis[f'e{i}'] = variable.Variable(name=f'e{i}', values=['Fail', 'Survive'])

    od_pair = ('n1', 'n3')
    edges = {
         'e1': {'origin': 'n1', 'destination': 'n2', 'link_capacity': None, 'weight': 1.0},
         'e2': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         'e3': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         }

    G = nx.MultiDiGraph()
    G.add_edge('n1', 'n2', label='e1', key='e1', weight=1)
    G.add_edge('n2', 'n3', label='e2', key='e2', weight=1)
    G.add_edge('n2', 'n3', label='e3', key='e3', weight=1)
    G.add_node('n1', key='n1', label='n1')
    G.add_node('n2', key='n2', label='n2')
    G.add_node('n3', key='n3', label='n3')

    s_path_edges, s_path_nodes = trans.get_connectivity_given_comps(x_star, G, od_pair)

    assert s_path_edges == ['e1', 'e3']
    assert s_path_nodes == ['n1', 'n2', 'n3']

    #sys_fun = trans.sys_fun_wrap(od_pair, edges, varis)

    #sys_val, sys_st, comp_st_min = sys_fun(x_star)

    #assert sys_st == 's'
    #assert sys_val == None
    #assert comp_st_min == {'e1': 1, 'e3': 1}


@pytest.mark.skip('NYI')
def test_get_csys2(main_sys_bridge):

    od_pair, arcs, varis = main_sys_bridge
    comps_name = list(arcs.keys())

    # Intact state of component vector: zero-based index 
    comps_st_itc = {k: len(v.values) - 1 for k, v in varis.items()} # intact state (i.e. the highest state)
    d_time_itc, path_itc = trans.get_time_and_path_given_comps(comps_st_itc, od_pair, arcs, varis)

    # defines the system failure event
    thres = 2

    # Given a system function, i.e. sf_min_path, it should be represented by a function that only has "comps_st" as input.
    # can we define multi state for system here??
    sys_fun = trans.sys_fun_wrap(od_pair, arcs, varis, thres * d_time_itc)

    # Branch and bound
    output_path = Path(__file__).parent
    brs, rules, _, _ = brc.run(probs, sys_fun, max_sf=1000, max_nb=1000),


    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}
    #pdb.set_trace()
    result = brc.get_csys(brs, varis, st_br_to_cs)
    #cmat = result[0]
    cmat = result[0][np.argsort(result[0][:, 0])]
    expected = np.array([[0, 4, 0, 4, 3, 3, 3],
                         [0, 4, 0, 2, 3, 4, 3],
                         [0, 3, 0, 2, 3, 2, 3],
                         [0, 4, 3, 3, 4, 0, 3],
                         [0, 4, 3, 3, 2, 0, 0],
                         [0, 4, 1, 3, 2, 0, 1],
                         [1, 2, 0, 2, 3, 2, 3],
                         [1, 4, 2, 3, 2, 0, 1],
                         [1, 4, 3, 3, 2, 0, 2],
                         [1, 4, 3, 3, 3, 5, 3]])
    #print(varis['e1'])
    #print(varis['e3'])
    # FIXME: not exactly matching
    #np.testing.assert_array_equal(cmat, expected)
    """
    np.testing.assert_array_equal(result[1]['e1'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0], [1, 1, 1]]))
    np.testing.assert_array_equal(result[1]['e2'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1]]))
    np.testing.assert_array_equal(result[1]['e3'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1], [1, 1, 0]]))
    np.testing.assert_array_equal(result[1]['e4'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1], [1, 1, 0]]))
    np.testing.assert_array_equal(result[1]['e5'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1], [1, 1, 0], [0, 1, 1]]))
    np.testing.assert_array_equal(result[1]['e6'].B, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1], [1, 1, 0]]))
    """


@pytest.mark.skip('NYI')
@pytest.fixture()
def main_sys_bridge(data_bridge):

    # Network
    _, arcs, arc_times_h = info_bridge

    # od_pair is different from setup_bridge in test_trans.py
    od_pair = ('n5', 'n1')

    #arc_lens_km = trans.get_arcs_length(arcs, node_coords)
    #arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    #comps_name = [k for k in arcs] # *FIXME* this is not necessary if branch's down and up are defined by dictionary (instead of list)

    # Component events
    #no_arc_st = 2 # number of component states 
    #delay_rat = [10, 2, 1] # delay in travel time given each component state (ratio)
    varis = {}
    for k, v in arcs.items():
        #varis[k] = variable.Variable(name=k, B=[{0}, {1}, {2}] , values = [arc_times_h[k]*np.float64(x) for x in delay_rat])
        varis[k] = variable.Variable(name=k, values = [10.0*arc_times_h[k], arc_times_h[k]])

    return od_pair, arcs, varis


@pytest.fixture()
def comps_st_dic():

    comps_st = {}
    expected = {}

    # 0, 1, 2 (higher, better)
    # intact state (i.e. the highest state)
    comps_st[0] = {'e1': 2, 'e2': 2, 'e3': 2, 'e4': 2, 'e5': 2, 'e6': 2}
    expected[0] = (0.1844, 's', {'e2': 2, 'e5': 2}, ['n1', 'n5', 'n3'])

    comps_st[1] = {'e1': 0, 'e2': 1, 'e3': 0, 'e4': 0, 'e5': 0, 'e6': 2}
    expected[1] = (1.12308, 'f', {}, ['n1', 'n5', 'n3'])

    comps_st[2] = {'e1': 2, 'e2': 0, 'e3': 2, 'e4': 2, 'e5': 2, 'e6': 2}
    expected[2] = (0.3344, 's', {'e1': 2, 'e3': 2 , 'e5': 2}, ['n1', 'n2', 'n5', 'n3'])

    comps_st[3] = {'e1': 0, 'e2': 2, 'e3': 1, 'e4': 2, 'e5': 1, 'e6': 1}
    expected[3] = (0.2787, 's', {'e2': 2, 'e5': 1}, ['n1', 'n5', 'n3'])

    comps_st[4] = {'e1': 2, 'e2': 2, 'e3': 0, 'e4': 1, 'e5': 1, 'e6': 2}
    expected[4] = (0.2787, 's', {'e2': 2, 'e5': 1}, ['n1', 'n5', 'n3'])

    comps_st[5] = {'e1':2, 'e2':1, 'e3':2, 'e4':2, 'e5':2, 'e6': 2}
    expected[5] = (0.2746, 's', {'e2': 1, 'e5': 2}, ['n1', 'n5', 'n3'])

    comps_st[6] = {'e1':2, 'e2':2, 'e3':2, 'e4':2, 'e5':1, 'e6': 2}
    expected[6] = (0.26626, 's', {'e2': 2, 'e6': 2, 'e4':2}, ['n1', 'n5', 'n4', 'n3'])

    comps_st[7] = {'e1':1, 'e2':0, 'e3':2, 'e4':2, 'e5':2, 'e6': 2}
    expected[7] = (0.4844, 'f', {}, ['n1', 'n2', 'n5', 'n3'])

    comps_st[7] = {'e1':2, 'e2':0, 'e3':1, 'e4':2, 'e5':2, 'e6': 2}
    expected[7] = (0.4246, 'f', {}, ['n1', 'n2', 'n5', 'n3'])


    return comps_st, expected


def test_proposed_branch_and_bound_using_probs(main_sys):
    # Branch and bound
    G, od_pair, arcs, d_varis, _ = main_sys

    varis = copy.deepcopy(d_varis)

    # Intact state of component vector: zero-based index 
    comps_st_itc = {k: len(v.values) - 1 for k, v in varis.items()} # intact state (i.e. the highest state)
    d_time_itc, path_n, path_itc = trans.get_time_and_path_given_comps(comps_st_itc, G, od_pair, varis)

    # defines the system failure event
    thres = 2

    # Given a system function, i.e. sf_min_path, it should be represented by a function that only has "comps_st" as input.
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres * d_time_itc)

    p = [1/3, 1/3, 1/3]
    probs = {'e1': p, 'e2': p, 'e3': p,
             'e4': p, 'e5': p, 'e6': p}

    # Branch and bound
    output_path = Path(__file__).parent
    #t1 = time.perf_counter()
    brs, rules, _, monitor = brc.run(probs, sys_fun, max_sf=100, max_nb=100)

    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}
    csys, varis = brc.get_csys(brs, varis, st_br_to_cs)

    varis['sys'] = variable.Variable('sys', ['f', 's', 'u'])
    cpm_sys_vname = list(brs[0].up.keys())
    cpm_sys_vname.insert(0, 'sys')

    # Component events
    cpms = {}
    for k, v in arcs.items():
        cpms[k] = cpm.Cpm(variables=[varis[k]],
                          no_child = 1,
                          C = np.array([0, 1, 2]),
                          p = [0.1, 0.2, 0.7])

    cpms['sys'] = cpm.Cpm(
        variables=[varis[k] for k in cpm_sys_vname],
        no_child = 1,
        C = csys,
        p = np.ones((len(csys), 1), int))

    var_elim_order = [varis[k] for k in arcs.keys()]
    Msys = inference.variable_elim([cpms[v] for v in varis.keys()], var_elim_order )
    np.testing.assert_array_almost_equal(Msys.C, np.array([[0, 1]]).T)
    np.testing.assert_array_almost_equal(Msys.p, np.array([[0.1018, 0.8982]]).T)


def test_get_csys(setup_brs):

    varis, brs = setup_brs

    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}
    csys, varis = brc.get_csys(brs, varis, st_br_to_cs)

    for i in range(1, 7):
        assert compare_list_of_sets(varis[f'e{i}'].B, [{0}, {1}, {2}, {0, 1}, {0, 2}, {1, 2}, {0, 1, 2}])
    #assert compare_list_of_sets(varis['e1'].B, [{0}, {1}, {2}, {0, 1}, {0, 1, 2}])
    #assert compare_list_of_sets(varis['e2'].B, [{0}, {1}, {2}, {1, 2}])
    #assert compare_list_of_sets(varis['e3'].B, [{0}, {1}, {2}, {0, 1}, {0, 1, 2}])
    #assert compare_list_of_sets(varis['e4'].B, [{0}, {1}, {2}, {0, 1, 2}, {0, 1}])
    #assert compare_list_of_sets(varis['e5'].B, [{0}, {1}, {2}, {0, 1, 2}, {0, 1}, {1, 2}])
    #assert compare_list_of_sets(varis['e6'].B, [{0}, {1}, {2}, {0, 1, 2}, {0, 1}])

    expected = np.array([[0, 3, 0, 6, 6, 6, 6],
                         [0, 2, 0, 3, 6, 6, 6],
                         [0, 2, 0, 2, 6, 3, 6],
                         [0, 6, 5, 6, 3, 0, 6],
                         [0, 6, 1, 6, 2, 0, 3],
                         [0, 6, 2, 6, 2, 0, 0],
                         [1, 2, 0, 2, 6, 2, 6],
                         [1, 6, 2, 6, 2, 0, 1],
                         [1, 6, 5, 6, 2, 0, 2],
                         [1, 6, 5, 6, 6, 5, 6]])

    assert compare_list_of_sets(csys, expected)


def test_get_csys3(main_sys):

    G, od_pair, arcs, d_varis, _ = main_sys

    varis = copy.deepcopy(d_varis)

    p1 = {0: 0.05, 1: 0.15, 2: 0.80}
    p2 = {0: 0.01, 1: 0.09, 2: 0.90}

    probs = {'e1': p1, 'e2': p1, 'e3': p1,
             'e4': p2, 'e5': p2, 'e6': p2}

    #comps_name = list(arcs.keys())

    # Intact state of component vector: zero-based index 
    comps_st_itc = {k: len(v.values) - 1 for k, v in varis.items()} # intact state (i.e. the highest state)
    d_time_itc, path_n, path_itc = trans.get_time_and_path_given_comps(comps_st_itc, G, od_pair, varis)

    # defines the system failure event
    thres = 2

    # Given a system function, i.e. sf_min_path, it should be represented by a function that only has "comps_st" as input.
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis, thres * d_time_itc)

    # Branch and bound
    output_path = Path(__file__).parent
    #pdb.set_trace()
    brs, rules, _, monitor = brc.run(probs, sys_fun, max_sf=100, max_nb=100)

    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}
    #pdb.set_trace()
    csys, varis = brc.get_csys(brs, varis, st_br_to_cs)

    varis['sys'] = variable.Variable('sys', ['f', 's', 'u'])
    cpm_sys_vname = list(brs[0].up.keys())
    cpm_sys_vname.insert(0, 'sys')

    # Component events
    cpms = {}
    for k, v in arcs.items():
        cpms[k] = cpm.Cpm(variables=[varis[k]],
                          no_child = 1,
                          C = np.array([0, 1, 2]),
                          p = [0.1, 0.2, 0.7])

    cpms['sys'] = cpm.Cpm(
        variables=[varis[k] for k in cpm_sys_vname],
        no_child = 1,
        C = csys,
        p = np.ones((len(csys), 1), int))

    var_elim_order = [varis[k] for k in arcs.keys()]
    Msys = inference.variable_elim([cpms[v] for v in varis.keys()], var_elim_order )
    np.testing.assert_array_almost_equal(Msys.C, np.array([[0, 1]]).T)
    np.testing.assert_array_almost_equal(Msys.p, np.array([[0.1018, 0.8982]]).T)


def test_get_connectivity_given_comps4():

    x_star = {'e1': 1, 'e2': 0, 'e3': 1}
    varis = {}
    for i in range(1, 4):
        varis[f'e{i}'] = variable.Variable(name=f'e{i}', values=['Fail', 'Survive'])

    od_pair = ('n1', 'n3')
    edges = {
         'e1': {'origin': 'n1', 'destination': 'n2', 'link_capacity': None, 'weight': 1.0},
         'e2': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         'e3': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         }

    G = nx.MultiDiGraph()
    G.add_edge('n1', 'n2', label='e1', key='e1', weight=1)
    G.add_edge('n2', 'n3', label='e2', key='e2', weight=1)
    G.add_edge('n2', 'n3', label='e3', key='e3', weight=1)
    G.add_node('n1', key='n1', label='n1')
    G.add_node('n2', key='n2', label='n2')
    G.add_node('n3', key='n3', label='n3')

    s_path_edges, s_path_nodes = trans.get_connectivity_given_comps(x_star, G, od_pair)

    assert s_path_edges == ['e1', 'e3']
    assert s_path_nodes == ['n1', 'n2', 'n3']

    #sys_fun = trans.sys_fun_wrap(od_pair, edges, varis)

    #sys_val, sys_st, comp_st_min = sys_fun(x_star)

    #assert sys_st == 's'
    #assert sys_val == None
    #assert comp_st_min == {'e1': 1, 'e3': 1}




import pytest


from demos.pf_map import batch as pf_map
from demos.rbd import batch as rbd
from demos.power_house import batch as power_house
from demos.routine import batch as routine
from demos.road import batch as road
from demos.SF import batch as sf


def test_batch():

    #rbd.main()
    pf_map.debug()
    #routine.main()
    #road.main()
    #sf.main(max_sf=10)
    #power_house.batch()



import numpy as np
import copy
import pytest

from mbnpy import bnb_fns, branch, cpm, inference


@pytest.mark.skip('run_bnb')
def test_bnb(setup_bridge, expected_probs):

    d_cpms_arc, d_vars_arc, arcs, var_ODs = setup_bridge

    cpms_arc = copy.deepcopy(d_cpms_arc)
    vars_arc = copy.deepcopy(d_vars_arc)

    ## Problem
    info = {'path': [['e2'], ['e3', 'e1']],
            'time': np.array([0.0901, 0.2401]),
            'arcs': list(arcs.keys()),
            'max_state': 2
            }

    max_state = 2
    comp_max_states = (max_state*np.ones(len(arcs))).tolist()

    # run_bnb not working
    branches = branch.run_bnb(sys_fn=bnb_fns.bnb_sys,
                       next_comp_fn=bnb_fns.bnb_next_comp,
                       next_state_fn=bnb_fns.bnb_next_state,
                       info=info,
                       comp_max_states=comp_max_states)

    C_od= branch.get_cmat(branches, [vars_arc[i] for i in arcs.keys()])

    # Check if the results are correct
    # FIXME: index issue
    od_var_id = 'od1' #7 - 1
    var_elim_order = [vars_arc[i] for i in arcs.keys()]
    #pdb.set_trace()
    #M_bnb = {i: cpms_arc[i] for i in list(arcs.keys()) + list(var_ODs.keys())}
    M_bnb = {i: cpms_arc[i] for i in list(arcs.keys()) + ['od1']}
    #M_bnb = [cpms_arc[i] for i in list(arcs.keys()) + ['od1']]
    #M_bnb[od_var_id].C = C_od
    #M_bnb[od_var_id].p = np.ones(shape=(C_od.shape[0], 1))
    M_bnb_VE= inference.variable_elim(M_bnb, var_elim_order)

    #print(M_bnb_VE)
    # FIXME: index issue
    disconn_state = 0 # max basic state
    disconn_prob = M_bnb_VE.get_prob([vars_arc['od1']], [disconn_state])
    delay_prob = M_bnb_VE.get_prob([vars_arc['od1']], [0]) + M_bnb_VE.get_prob([vars_arc['od1']], [1])

    # Check if the results are the same
    # FIXME: index issue
    np.testing.assert_array_almost_equal(expected_probs['disconn'][0], disconn_prob, decimal=4)
    np.testing.assert_array_almost_equal(expected_probs['delay'][0], delay_prob, decimal=4)

    # using variable name instead
    disconn_state = 0 # max basic state
    disconn_prob = M_bnb_VE.get_prob(['od1'], np.array([disconn_state]))
    delay_prob = M_bnb_VE.get_prob(['od1'], [0]) + M_bnb_VE.get_prob(['od1'], [1])
    #delay_prob = cpm.get_prob(M_bnb_VE, ['od1'], np.array([1-1]), 0)

    np.testing.assert_array_almost_equal(expected_probs['disconn'][0], disconn_prob, decimal=4)
    np.testing.assert_array_almost_equal(expected_probs['delay'][0], delay_prob, decimal=4)






import pandas as pd
import pdb
import typing
import copy
from pathlib import Path
from collections import Counter
from itertools import chain
import matplotlib.pyplot as plt
import warnings
import sys
import pickle
import numpy as np
import time
import json
#import editdistance

from mbnpy import variable, branch


def run(probs, sys_fun, rules=None, brs=None, max_sf=np.inf, max_nb=np.inf, pf_bnd_wr=0.0, max_rules=100, surv_first=True, active_decomp=10, final_decomp=True, display_freq=10, autosave=50):

    """
    Run the BRC algorithm to find (1) non-dominated rules and
    (2) branches for system reliability analysis.

    Attributes:
        probs (dictionaty): {comp_name (str): probabilities (list)}
        sys_fun: a system function
            One attribute:
                comp_state (dictionary): {comp_name (str): state (int)}
            Returns (orders need to be kept):
                system value (any type)
                system state ('s' or 'f')
                minimum rule for system state (dictionary):
                    {comp_name (str): state (int)}

        **Information from previous analysis (optional when available)**
            rules (dictionary): {'s': list of rules, 'f': list of rules}
            brs (list): branches from previous analysis

        **Iteration termination conditions**
            max_sf (int): maximum number of system function runs
            max_nb (int): maximum number of branches
            pf_bnd_wr (float, non-negative): bound of system failure probability
                in ratio (width / lower bound)
            max_rules (int): the maximum number of rules
        **Decomposition options**
            surv_first: True if survival branches are considered first
            active_decomp: True if branches are re-obtained at each iteration
                False if branches are never re-evaluated
                int (>0) if regularly re-evaluated
            final_decomp: True if final decomposition is performed
                (only when active_decomp = False)
        **Display options**
            display_freq (int): frequency of displaying the current progress
        **Autosave options**
            autosave (int): frequency of saving the current progress
                (brs, rules, sys_res, monitor) to files

    Returns:
        brs (list): branches
        rules (dictionary): {'s': list of rules, 'f': list of rules}
        sys_res (pandas.DataFrame): system function results
        monitor (dictionary): monitoring information
    """

    if not rules:
        rules = {'s': [], 'f': []}

    sys_res = pd.DataFrame(data={'sys_val': [],
                                 'comp_st': [],
                                 'comp_st_min': []}) # system function results

    monitor, ctrl = init_monitor()

    while ctrl['no_sf'] < max_sf:

        start = time.time() # monitoring purpose

        if active_decomp is True:
            brs, _ = decomp_depth_first(rules, probs, max_nb)  # S2
        elif active_decomp is False or active_decomp==0:

            if brs is None:
                brs = []
            brs, _ = decomp_depth_first(rules, probs, max_nb, brs) # existing branches are not reassessed.
        else:
            if ctrl['no_sf'] % active_decomp == 0:
                brs, _ = decomp_depth_first(rules, probs, max_nb)  # S2

            else:
                if brs is None:
                    brs = []
                brs, _ = decomp_depth_first(rules, probs, max_nb, brs) # existing branches are not reassessed.

        x_star = get_comp_st(brs, surv_first, probs)  # S4-1

        if x_star == None:
            monitor['out_flag'] = 'complete'
            break

        elif len(brs) >= max_nb:
            monitor['out_flag'] = 'max_nb'
            break

        elif ctrl['pr_bu'] < ctrl['pf_low'] * pf_bnd_wr:
            monitor['out_flag'] = 'pf_bnd'
            break

        elif len(rules['s']+rules['f']) >= max_rules:
            monitor['out_flag'] = 'max_rules'
            break

        else:
            rule, sys_res_ = run_sys_fn(x_star, sys_fun, probs) # S4-2, S5

            rules = update_rule_set(rules, rule) # S6
            sys_res = pd.concat([sys_res, sys_res_], ignore_index=True)
            ctrl['no_sf'] += 1

            monitor['no_sf'].append(ctrl['no_sf'])
            monitor, ctrl = update_monitor(monitor, brs, rules, start) # S7

            if ctrl['no_sf'] % display_freq == 0:
                print(f"[System function runs {ctrl['no_sf']}]..")
                display_msg(monitor)

            if autosave and ctrl['no_sf'] % autosave == 0:
                save_brc_data(rules, brs, sys_res, monitor, fname_suffix="autosave")

        if ctrl['no_sf'] == max_sf:
            monitor['out_flag'] = 'max_sf'

    try:
        if final_decomp and (active_decomp is False or active_decomp > 1):
            nbr_old = len(brs)
            brs, _ = decomp_depth_first(rules, probs, max_nb)
            print(f"\n*Final decomposition is completed with {len(brs)} branches (originally {nbr_old} branches).")

        monitor, ctrl = update_monitor(monitor, brs, rules, start)

        print(f"\n***Analysis completed with f_sys runs {ctrl['no_sf']}: out_flag = {monitor['out_flag']}***")
        display_msg(monitor)

    except NameError: # analysis is terminated before the first system function run
        print(f'\n***Analysis terminated without any evaluation***')

    return brs, rules, sys_res, monitor


def init_monitor():

    monitor = {'pf_up': [], # upper bound on pf
               'pf_low': [], # lower bound on pf
               'pr_brs': [], # prob. of branches (can be <1 in case event space is a subset of total space)
               'pr_bu': [], # prob. of unknown branches
               'no_br': [], # number of branches
               'no_bs': [], # number of branches-survival (br_s_ns)
               'no_bf': [], # number of branches-failure (br_f_ns)
               'no_bu': [], # number of branches-unknown (br_us_ns)
               'no_rs': [], # number of rules-survival (r_s_ns)
               'no_rf': [], # number of rules-failure (r_f_ns)
               'no_ra': [], # number of rules (no_rs + no_rf)
               'no_sf': [0], # number of system function runs (sf_ns)
               'time': [], # time taken for each iteration (sec)
               'min_len_rf': [], # min. length of rules-failure
               'avg_len_rf': [], # avg. length of rules-failure
               'out_flag': None, # outflag ('complete', 'max_nb', 'pf_bnd', 'max_sf'),
               'max_bu': 0 # max. number of branches-unknown 
              }

    # init for ctrl
    # no_sf, pr_bf, pr_bs = 0, 0, 0
    ctrl = {'no_sf': 0, # no_sf
            'pr_bu': 1, # 1 - pr_bf - pr_bs
            'pf_low': 0, # pr_bf
            'no_bu': 1, # no_bu
            }

    return monitor, ctrl


def update_monitor(monitor, brs, rules, start):

    end = time.time() # monitoring purpose

    pr_brs = sum([br.p for br in brs]) # prob. of branches (<1 in case event space is a subset of total space)
    pr_bf = sum([br.p for br in brs if br.up_state == 'f']) # prob. of failure branches
    pr_bs = sum([br.p for br in brs if br.down_state == 's']) # prob. of survival branches
    pr_bu = pr_brs - pr_bf - pr_bs # prob. of unknown branches

    monitor['pf_low'].append(pr_bf) # lower bound on pf
    monitor['pf_up'].append(pr_brs - pr_bs)  # upper bound of pf
    monitor['pr_bu'].append(pr_brs - pr_bf - pr_bs) # prob. of unknown branches
    monitor['pr_brs'].append(pr_brs) # prob. of branches 

    no_rf = len(rules['f'])
    no_rs = len(rules['s'])

    monitor['no_rf'].append(no_rf)
    monitor['no_rs'].append(no_rs)
    monitor['no_ra'].append(no_rs + no_rf)

    no_br = len(brs)
    no_bf = sum([b.up_state == 'f' for b in brs])
    no_bs = sum([b.down_state == 's' for b in brs])
    no_bu = no_br - no_bf - no_bs

    monitor['no_bf'].append(no_bf)
    monitor['no_bs'].append(no_bs)
    monitor['no_bu'].append(no_bu)
    monitor['no_br'].append(no_br)

    monitor['time'].append(end - start)

    try:
        min_len_rf = min([len(x) for x in rules['f']])
        avg_len_rf = sum([len(x) for x in rules['f']]) / no_rf
    except ValueError:
        min_len_rf = 0
        avg_len_rf = 0

    monitor['min_len_rf'].append(min_len_rf)
    monitor['avg_len_rf'].append(avg_len_rf)

    # get the latest value for ctrl
    keys = ['no_sf', 'pr_bu', 'pf_low', 'no_bu']
    ctrl = {}
    for k in keys:
        try:
            ctrl[k] = monitor[k][-1]
        except TypeError:
            ctrl[k] = monitor[k]

    return monitor, ctrl


def display_msg(monitor):

    last = {}
    for k, v in monitor.items():
        try:
            last[k] = v[-1]
        except TypeError:
            last[k] = v

    print(f"The # of found non-dominated rules (f, s): {last['no_ra']} ({last['no_rf']}, {last['no_rs']})")
    print(f"Probability of branchs (f, s, u): ({last['pf_low']:.4e}, {last['pr_brs']-last['pf_up']:.2e}, {last['pr_bu']:.4e})")
    print(f"The # of branches (f, s, u), (min, avg) len of rf: {last['no_br']} ({last['no_bf']}, {last['no_bs']}, {last['no_bu']}), ({last['min_len_rf']}, {last['avg_len_rf']:.2f})")
    print(f"Elapsed seconds (average per round): {sum(monitor['time']):1.2e} ({np.mean(monitor['time']):1.2e})")


def plot_monitoring(monitor, output_file='monitor.png'):
    """

    """

    # bounds vs no. of branches
    fig = plt.figure(figsize=(6, 4*3))
    ax = fig.add_subplot(311)
    ax.plot(monitor['no_br'], monitor['pf_low'], linestyle='--', color='blue')
    ax.plot(monitor['no_br'], monitor['pf_up'], linestyle='--', color='blue')
    ax.set_xlabel('No. of branches')
    ax.set_ylabel('System failure prob. bounds')

    # bounds vs sys fn runs
    ax = fig.add_subplot(312)
    ax.plot(monitor['no_sf'], monitor['pf_low'], linestyle='--', color='blue')
    ax.plot(monitor['no_sf'], monitor['pf_up'], linestyle='--', color='blue')
    ax.set_xlabel('No. of system function runs')
    ax.set_ylabel('System failure prob. bounds')

    # no. of rules vs sys fn runs
    ax = fig.add_subplot(313)
    ax.plot(monitor['no_ra'], monitor['pf_low'], linestyle='--', color='blue')
    ax.plot(monitor['no_ra'], monitor['pf_up'], linestyle='--', color='blue')
    ax.set_xlabel('No. of rules')
    ax.set_ylabel('System failure prob. bounds')

    #output_file = Path(sys.argv[1]).joinpath(output_file)
    fig.savefig(output_file, dpi=200)
    print(f'{output_file} created')


def get_csys(brs, varis, st_br_to_cs):
    """

    """
    c_sys = np.empty(shape=(0, len(brs[0].up.keys()) + 1), dtype=int)

    for br in brs:
        varis, c = br.get_c(varis, st_br_to_cs)
        c_sys = np.vstack([c_sys, c])

    return c_sys, varis


def get_state(comp, rules):
    """
    Args:
        comp (dict): component state vector in dictionary
                     e.g., {'x1': 0, 'x2': 0 ... }
        rules (list): a list of rules
                     e.g., {({'x1': 2, 'x2': 2}, 's')}
    Returns:
        str: system state ('s', 'f', or 'u')
    """
    assert isinstance(comp, dict), f'comp should be a dict: {type(comp)}'
    assert isinstance(rules, dict), f'rules should be a dict: {type(rules)}'

    # the survival rule is satisfied
    s_rules = [i for (i,rule) in enumerate(rules['s']) if all([comp[k] >= v for k, v in rule.items()])]
    no_s = len(s_rules)

    f_rules = [i for (i,rule) in enumerate(rules['f']) if all([comp[k] <= v for k, v in rule.items()])]
    no_f = len(f_rules)

    # no compatible rules. the state remains unknown
    if no_s == no_f == 0:
        state = 'u'
    elif no_s >= no_f:
        state = 's'
    else:
        state = 'f'

    if no_s > 0 and no_f > 0:
        rules_s = [rules['s'][i] for i in s_rules]
        rules_f = [rules['f'][i] for i in f_rules]

        print(f"Conflicting rules found: {rules_s} vs. {rules_f}. The given system is not coherent.")

    return state


def update_rule_set(rules, new_rule):
    """
    rules: dict of rule list
           e.g., {'s': [{'x1': 2, 'x2': 2}, ..], 'f': [{..}, ..]}
    new_rule: tutple of (rule, system state)
             e.g., ({'x1': 2}, 's')
    """
    assert isinstance(new_rule, tuple), f'rule should be a tuple: {type(new_rule)}'
    add_rule = True

    n_rule, n_state = new_rule
    #rules_old = copy.deepcopy(rules)

    if n_state == 's':

        for rule in rules['s'][:]:

            if set(n_rule).issubset(rule) and all([rule[k] >= v for k, v in n_rule.items()]):
                rules['s'].remove(rule)

            elif set(rule).issubset(n_rule) and all([n_rule[k] >= v for k, v in rule.items()]):
                add_rule = False
                break

    elif n_state == 'f':

        for rule in rules['f'][:]:

            if set(n_rule).issubset(rule) and all([rule[k] <= v for k, v in n_rule.items()]):
                rules['f'].remove(rule)

            elif set(rule).issubset(n_rule) and all([n_rule[k] <= v for k, v in rule.items()]):
                add_rule = False
                break

    if add_rule:
        rules[n_state].append(n_rule)

    return rules


def run_sys_fn(comp: dict, sys_fun: typing.Callable, probs: dict):
    """
    comp: component vector state in dictionary
    e.g., {'x1': 0, 'x2': 0, ... }
    sys_fun
    rules: list of rules
           e.g., {({'x1': 2, 'x2': 2}, 's')}
    """

    assert isinstance(comp, dict), f'comp should be a dict: {type(comp)}'

    # S4-2: get system state given comp
    sys_val, sys_st, comp_st_min = sys_fun(comp)

    sys_res = pd.DataFrame({'sys_val': [sys_val], 'comp_st': [comp], 'comp_st_min': [comp_st_min]})

    if comp_st_min:
        rule = comp_st_min, sys_st

    else:
        if sys_st == 's':
            rule = {k: v for k, v in comp.items() if v}, sys_st # the rule is the same as up_dict but includes only components whose state is greater than the worst one (i.e. 0)
        else:
            rule = {k: v for k, v in comp.items() if v < len(probs[k]) - 1}, sys_st # the rule is the same as up_dict but includes only components whose state is less than the best one

    return rule, sys_res


def init_branch(probs: dict, rules: dict) -> list:
    """
    initialise a branch set (x_min, x_max, s(x_min), s(x_max), 1)
    """
    assert isinstance(probs, dict)
    assert isinstance(rules, dict)

    down = {x: 0 for x in probs.keys()}
    up = {k: len(v) - 1 for k, v in probs.items()}

    down_state = get_state(down, rules)
    up_state = get_state(up, rules)

    return [branch.Branch(down, up, down_state, up_state, 1.0)]


def decomp_depth_first(rules, probs, max_nb=np.inf, brs = []):
    """
    depth-first decomposition of event space using given rules
    """

    if len(brs) < 1:
        brs = init_branch(probs, rules)  # D1
    #crules = [brs[0].get_compat_rules(rules)]
    crules = [br.get_compat_rules(rules) for br in brs]

    go = True
    while go:

        # D2: sort branches from higher to lower p 
        sorted_brs = sorted(zip(brs, crules), key=lambda x: x[0].p, reverse=True)
        brs, crules = [list(x) for x in zip(*sorted_brs)]

        brs_new = []
        crules_new = []

        for i, (br, cr) in enumerate(sorted_brs, 1):

            br.eval_state(rules)

            # specified branch or no compatible rule exists
            if ((br.down_state == br.up_state) and (br.down_state != 'u')) or (len(cr['f']) + len(cr['s']) == 0):
                brs_new.append(br)
                crules_new.append({'s':[], 'f':[]})

            else:
                # D6??
                xd, xd_st = br.get_decomp_comp_using_probs(cr, probs)

                # D3: evaluate sl and su
                for up_flag in [True, False]:
                    br_new = br.get_new_branch(rules, probs, xd, xd_st, up_flag)
                    crule_new = br_new.get_compat_rules(rules)
                    brs_new.append(br_new)
                    crules_new.append(crule_new)

                #n_br = len(brs_new) + len(brs) - i # the current number of branches

                if len(brs_new) + len(brs) > max_nb:

                    go = False
                    brs_new += brs[i:]
                    crules_new += crules[i:]
                    break

        brs = copy.deepcopy(brs_new)
        crules = copy.deepcopy(crules_new)

        if go and sum([len(r['f']) + len(r['s']) for r in crules]) == 0:
           go = False

    return brs, crules


def get_comp_st(brs, surv_first=True, probs=None):
    """
    get a component vector state from branches(brs)
    'brs' is a list of branches obtained by depth-first decomposition
    """

    if surv_first:

        brs = sorted(brs, key=lambda x: x.p, reverse=True)

        x_star = None
        for br in brs: # look at up_state first
            if br.up_state == 'u':
                x_star = br.up
                break

        if x_star == None:
            for br in brs: # if all up states are known then down states
                if br.down_state == 'u':
                    x_star = br.down
                    break

    else:

        worst = {x: 0 for x in probs.keys()}
        best = {k: len(v.keys()) - 1 for k, v in probs.items()}

        brs_new = []
        for br in brs:
            if br.up_state == 'u':
                p_new = branch.approx_prob_by_comps(br.up, best, probs)
                br_new = branch.Branch(br.up, best, 'u', 's', p_new)
                brs_new.append(br_new)

            if br.down_state =='u':
                p_new = branch.approx_prob_by_comps(worst, br.down, probs)
                b_new = branch.Branch(worst, br.down, 'f', 'u', p_new)
                brs_new.append(br_new)

        x_star = None
        if brs_new:
            brs_new = sorted(brs_new, key=lambda x: x.p, reverse=True)
            if brs_new[0].up_state == 'u':
                x_star = brs_new[0].up
            elif brs_new[0].down_state == 'u':
                x_star = brs_new[0].down

    return x_star


def run_MCS_indep_comps(probs, sys_fun, cov_t = 0.01):
    nsamp, nfail = 0, 0
    cov = 1.0
    while cov > cov_t:

        # generate samples
        nsamp += 1
        samp = {}
        for k, v in probs.items():
            st1 = np.random.choice(list(v.keys()), size=1, p=list(v.values()))
            samp[k] = st1[0]

        # run system function
        _, sys_st, _ = sys_fun(samp)

        if sys_st == 'f':
            nfail += 1

            pf = nfail / nsamp
            if nfail > 5:
                std = np.sqrt( pf*(1-pf) / nsamp )
                cov = std / pf

        if nsamp % 20000 == 0:
            print(f'nsamp: {nsamp}, cov: {cov}, pf: {pf}')

    return pf, cov, nsamp

def eval_rules_prob(rules_list, s_or_f, probs):

    probs_list = []

    if s_or_f == 'f': # rules_list contain failure rules
        for r in rules_list:
            p = 1.0
            for k, v in r.items():
                p *= sum([probs[k][x] for x in range(v+1)])

            probs_list.append(p)

    elif s_or_f == 's':
        max_st = {k: max(probs[k].keys()) for k in probs.keys()}

        for r in rules_list:
            p = 1.0
            for k, v in r.items():
                p *= sum([probs[k][x] for x in range(v, max_st[k]+1)])

            probs_list.append(p)
    else:
        raise ValueError('s_or_f input variable must be a string either "s" or "f".')

    return probs_list

def save_brc_data(rules, brs, sys_res, monitor, output_folder = "brc_data", fname_prefix="", fname_suffix = ""):
    if not Path(output_folder).exists():
        Path(output_folder).mkdir(parents=True, exist_ok=True)

    fpath_rule = f"./{output_folder}/{fname_prefix+'_' if fname_prefix else ''}rules{'_'+fname_suffix if fname_suffix else ''}.json"
    with open(fpath_rule, "w") as f:
        json.dump(rules, f, indent=4)

    fpath_br = f"./{output_folder}/{fname_prefix+'_' if fname_prefix else ''}brs{'_'+fname_suffix if fname_suffix else ''}.parquet"
    branch.save_brs_to_parquet(brs, fpath_br)

    fpath_mon = f"./{output_folder}/{fname_prefix+'_' if fname_prefix else ''}monitor{'_'+fname_suffix if fname_suffix else ''}.json"
    with open(fpath_mon, "w") as f:
        json.dump(monitor, f, indent=4)

    fpath_res = f"./{output_folder}/{fname_prefix+'_' if fname_prefix else ''}sys_res{'_'+fname_suffix if fname_suffix else ''}.json"
    sys_res.to_json( fpath_res, orient='records', lines=True )


import copy
import logging
import configparser
import pandas as pd
import json
import networkx as nx
from pathlib import Path

try: # Do not enforce the installation of graphviz
    import graphviz as gv
    GRAPHVIZ_AVAILABLE = True
except ImportError:
    GRAPHVIZ_AVAILABLE = False


template_nodes = {
    "component_type": None,
    "component_class": None,
    "cost_fraction": None,
    "node_type": None,
    "node_cluster": None,
    "operating_capacity": None,
    "pos_x": None,
    "pos_y": None,
    "damages_states_constructor": None
    }

template_edges = {
    "origin": None,
    "destination": None,
    "link_capacity": None,
    "weight": None,
    }

graph_types = ['Graph',
               'DiGraph',
               'MultiGraph',
               'MultiDiGraph']


class Config(object):
    """
    """
    REQ = ['MODEL_NAME', 'MAX_BRANCHES', 'CONFIGURATION_ID', 'OUTPUT_PATH']

    def __init__(self, file_cfg):
        """
        :param file_cfg: config file
        """

        with open(file_cfg, 'r') as f:
            data = json.load(f)

        assert all(x in data for x in self.REQ), f'{self.REQ} should be defined in the config file'

        HOME = Path(file_cfg).absolute().parent

        # process required
        self.file_model = HOME.joinpath(data['MODEL_NAME'])
        assert self.file_model.exists(), f'{self.file_model} does not exist'

        self.infra = read_model_from_json(self.file_model)

        # scenario can be empty
        if data['SCENARIO_NAME']:
            self.file_scenarios = HOME.joinpath(data['SCENARIO_NAME'])
            assert self.file_scenarios.exists(), f'{file_scenarios} does not exist'
            self.scenarios = read_scenarios_from_json(self.file_scenarios)
            self.no_ds = len(self.scenarios['damage_states'])
            data.pop('SCENARIO_NAME')
        else:
            print(f'scenario to be added later')

        self.key = data['CONFIGURATION_ID']

        self.max_branches = data['MAX_BRANCHES']

        self.output_path = HOME.joinpath(data['OUTPUT_PATH'])
        if not self.output_path.exists():
            self.output_path.mkdir(parents=True, exist_ok=True)
            print(f'{self.output_path} created')

        [data.pop(x) for x in self.REQ]
        self.data = data.copy()


def read_model_from_json(file_input):

    with open(file_input, 'r') as f:
        # ensure that damage states are ordered
        #model = json.load(f, object_pairs_hook=OrderedDict)
        model = json.load(f)

    # read node information
    nodes = {}
    for k, v in model['component_list'].items():
        nodes[k] = v

    edges = {}
    for k, v in model['node_conn_df'].items():
        edges[k] = v

    ODs={}
    for k, v in model['sysout_setup'].items():
        ODs[k] = {k1:v1 for k1, v1 in v.items() if v1}

    # create a graph
    try:
        gtype = model['graph_type']
    except KeyError:
        G = nx.Graph()
    else:
        assert gtype in graph_types, f'graph_type should be one of {graph_types}'
        if gtype.lower() == 'graph':
            G = nx.Graph()
        elif gtype.lower() == 'digraph':
            G = nx.DiGraph()
        elif gtype.lower() == 'multigraph':
            G = nx.MultiGraph()
        elif gtype.lower() == 'multidigraph':
            G = nx.MultiDiGraph()

    for k, v in edges.items():
        G.add_edge(v['origin'], v['destination'], label=k, weight=v['weight'], capacity=v['link_capacity'], key=k)

    for k, v in nodes.items():
        G.add_node(k, pos=(v['pos_x'], v['pos_y']), label=k, key=k)

    # create length attribute and time 
    return {'G': G, 'nodes': nodes, 'edges': edges, 'ODs': ODs}


def read_scenarios_from_json(file_scenarios):

    with open(file_scenarios, 'r') as f:
        # ensure that damage states are ordered
        scenarios = json.load(f)

    return scenarios


def convert_csv_to_json(df, template):
    """
    df:
    template:
    """
    data = {}
    for k, v in df.iterrows():
        template.update(**v)
        data[str(k)] = copy.deepcopy(template)

    return data


def dict_to_json(dict_pf, damage_states, filename=None):
    """
    to create a json file for scenario
    dict_pf: dict of prob. of failiure
    damage_states: list of strings
    """

    assert isinstance(dict_pf, dict), 'dict_pf should be a dict'
    assert isinstance(damage_states, list), 'damage_states should be a list'

    no_scenarios = len(dict_pf[next(iter(dict_pf))])
    df = []
    for i in range(no_scenarios):
        tmp = [(v[i][0], 1-v[i][0]) for _, v in dict_pf.items()]
        df.append(tmp)

    df = pd.DataFrame(df).T
    df['index'] = dict_pf.keys()
    df = df.set_index('index')
    df = df.rename({k: f's{k+1}' for k in range(no_scenarios)}, axis=1)
    json_str = json.loads(df.to_json())

    if filename:
        with open(filename, 'w') as w:
            json.dump({'damage_states': damage_states,
                       'scenarios': json_str}, w, indent=4)
            print(f'{filename} is written')


def networkx_to_graphviz(g):
    """Convert `networkx` graph `g` to `graphviz.Digraph`.

    @type g: `networkx.Graph` or `networkx.DiGraph`
    @rtype: `graphviz.Digraph`
    """
    if g.is_directed():
        h = gv.Digraph()
    else:
        h = gv.Graph()
    for u, d in g.nodes(data=True):
        h.node(str(u), label=d['label'])
    for u, v, d in g.edges(data=True):
        h.edge(str(u), str(v), label=f"{d['label']}({d['weight']})")
    return h


def plot_graphviz(G, outfile='graph'):

    h = networkx_to_graphviz(G)
    h.render(outfile, format='png', cleanup=True)
    print(f'{outfile}.png is created')









import numpy as np
import textwrap
import json
import time
import copy
import gc
from collections import Counter
from itertools import chain
import pdb
import itertools
from pathlib import Path
from collections import namedtuple
from scipy.stats import beta
import pandas as pd
import pyarrow
#import dask
#from dask.distributed import Client, worker_client, as_completed, get_client

#import dask.bag as db
from mbnpy import cpm, variable, trans, brc


#attr = ["down", "up", "down_state", "up_state", "p"]
#Branch = namedtuple("Branch", attr, defaults=(None,)*len(attr))

#Branch_p = namedtuple("Branch", ["down", "up", "down_state", "up_state", "p"])

def approx_prob_by_comps(down, up, probs):
    """

    """
    p = 1.0
    for k, v in down.items():
       p *= sum([probs[k][x] for x in range(v, up[k] + 1)])
    return p


class Branch(object):

    def __init__(self, down, up, down_state=None, up_state=None, p=None):

        assert isinstance(down, dict), 'down should be a dict'
        assert isinstance(up, dict), 'down should be a dict'
        assert len(down) == len(up), 'Vectors "down" and "up" must have the same length.'

        self.down = down
        self.up = up
        self.down_state = down_state
        self.up_state = up_state
        self.p = p

        #assert isinstance(is_complete, bool), '"is_complete" must be either true (or 1) or false (or 0)'

    def __repr__(self):
        #return textwrap.dedent(f"""\{self.__class__.__name__}(down={self.down}, up={self.up}, down_state={self.down_state}, up_state={self.up_state}, p={self.p}""")
        return (f"Branch(\n"
                f"  down={self.down},\n"
                f"  up={self.up},\n"
                f"  down_state='{self.down_state}',\n"
                f"  up_state='{self.up_state}',\n"
                f"  p={self.p}\n"
                f")")

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Branch):
            return all([self.down == other.down,
                        self.up == other.up,
                        self.down_state == other.down_state,
                        self.up_state == other.up_state,
                        self.p == other.p])

    def approx_prob(self, probs): # TODO: the naming is wrong, It should be "get_indep_prob" (the probability is not approximate when components are independent.)
        self.p = approx_prob_by_comps(self.down, self.up, probs)


    def get_compat_rules(self, rules):

        """
        lower: lower bound on component vector state in dictionary
               e.g., {'x1': 0, 'x2': 0 ... }
        upper: upper bound on component vector state in dictionary
               e.g., {'x1': 2, 'x2': 2 ... }
        rules: dict of rules
               e.g., {'s': [{'x1': 2, 'x2': 2}],
                      'f': [{'x1': 2, 'x2': 0}]}
        """
        assert isinstance(rules, dict), f'rules should be a dict: {type(rules)}'

        compat_rules = {'s': [], 'f': []}

        for rule in rules['s']:
            if all([self.up[k] >= v for k, v in rule.items()]):
                c_rule = {k: v for k, v in rule.items() if v > self.down[k]}
                if c_rule:
                    compat_rules['s'].append(c_rule)

        for rule in rules['f']:
            if all([self.down[k] <= v for k, v in rule.items()]):
                c_rule = {k: v for k, v in rule.items() if v < self.up[k]}
                if c_rule:
                    compat_rules['f'].append(c_rule)

        return compat_rules


    def eval_state( self, rules ):

        """
        lower: lower bound on component vector state in dictionary
               e.g., {'x1': 0, 'x2': 0 ... }
        upper: upper bound on component vector state in dictionary
               e.g., {'x1': 2, 'x2': 2 ... }
        rules: dict of rules
               e.g., {'s': [{'x1': 2, 'x2': 2}],
                      'f': [{'x1': 2, 'x2': 0}]}
        """
        assert isinstance(rules, dict), f'rules should be a dict: {type(rules)}'

        for rule in rules['s']:
            if self.up_state != 'u':
                break

            if all([self.up[k] >= v for k, v in rule.items()]):
                self.up_state = 's'


        for rule in rules['s']:
            if self.down_state != 'u':
                break

            if all([self.down[k] >= v for k, v in rule.items()]):
                self.down_state = 's'

        for rule in rules['f']:
            if self.up_state != 'u':
                break

            if all([self.up[k] <= v for k, v in rule.items()]):
                self.up_state = 'f'


        for rule in rules['f']:
            if self.down_state != 'u':
                break

            if all([self.down[k] <= v for k, v in rule.items()]):
                self.down_state = 'f'


    def approx_joint_prob_compat_rule(self, rule, rule_st, probs):
        assert isinstance(rule, dict), f'rule should be a dict: {type(rule)}'
        assert isinstance(rule_st, str), f'rule_st should be a string: {type(rule_st)}'
        assert isinstance(probs, dict), f'probs should be a dict: {type(probs)}'
        p = 1.0
        if rule_st == 's':
            for x, v in rule.items():
                p *= sum([probs[x][i] for i in range(v, self.up[x] + 1)])

        elif rule_st == 'f':
            for x, v in rule.items():
                p *= sum([probs[x][i] for i in range(self.down[x], v + 1)])

        return p


    def get_decomp_comp_using_probs(self, rules, probs):
        """
        rules: dict of rules
               e.g., {'s': [{'x1': 2, 'x2': 2}], 'f': [{'x3': 0}]}
        probs: dict
        """
        assert isinstance(rules, dict), f'rules should be a dict: {type(rules)}'
        assert isinstance(probs, dict), f'probs should be a dict: {type(probs)}'

        # get an order of component by their frequency in rules
        rules_st = [(k, x) for k, rule in rules.items() for x in rule]
        comps = Counter(chain.from_iterable([x[1] for x in rules_st]))
        comps = [x[0] for x in comps.most_common()]

        # get an order R by P  (higher to lower)
        if len(rules_st) > 1:
            rules_st = sorted(rules_st, key=lambda x: self.approx_joint_prob_compat_rule(x[1], x[0], probs), reverse=True)

        for st, rule in rules_st:
            for c in comps:
                state = None
                if c in rule and st == 's':
                    state = rule[c]

                elif c in rule and st == 'f':
                    state = rule[c] + 1

                try:
                    if self.down[c] < state and state <= self.up[c]:
                        return c, state

                except TypeError:
                    pass

    def get_new_branch(self, rules, probs, xd, xd_st, up_flag=True):
        """

        """
        if up_flag:
            up = self.up.copy()
            up[xd] = xd_st - 1
            up_st = brc.get_state(up, rules)

            down = self.down
            down_st = self.down_state

        else:
            up = self.up
            up_st = self.up_state

            down = self.down.copy()
            down[xd] = xd_st
            down_st = brc.get_state(down, rules)

        new_br = Branch(down=down, up=up, down_state=down_st, up_state=up_st)
        new_br.approx_prob(probs)

        return new_br

    def get_c(self, varis, st_br_to_cs):
        """
        return updated varis and state
        varis: a dictionary of variables
        st_br_to_cs: a dictionary that maps state in br to state in C matrix of a system event
        """
        names = list(self.up.keys())
        cst = np.zeros(len(names) + 1, dtype=int) # (system, compponents)

        if self.down_state == self.up_state:
            cst[0] = st_br_to_cs[self.down_state]
        else:
            cst[0] = st_br_to_cs['u']

        for i, x in enumerate(names):
            down = self.down[x]
            up = self.up[x]

            if up > down:
                states = set(range(down, up + 1))
                st = varis[x].get_state(states)
            else:
                st = up

            cst[i + 1] = st

        return varis, cst


    def to_dict(self):
        return {'down': self.down, 'up': self.up, 'down_state': self.down_state, 'up_state': self.up_state, 'p': self.p}

    @staticmethod
    def from_dict(data):
        return Branch( down = data['down'], up = data['up'], down_state = data['down_state'], up_state = data['up_state'], p = data['p'] )


def save_brs_to_parquet(brs_list, file_path):
    """
    Save a list of Branch objects to a .parquet file with efficient handling
    of nested dictionaries by flattening them into columns.

    Parameters:
    brs_list (list): A list of Branch objects.
    file_path (str): The full path to save the .parquet file.

    Example usage (given that `branches` is a list of Branch objects):
    save_branches_to_parquet(branches, 'output_file.parquet')

    """
    # Create a list to hold flattened data
    data = []

    # Iterate over the branches to create a flattened dictionary for each Branch
    for br in brs_list:
        # Flatten the 'down' and 'up' dictionaries by prefixing keys for uniqueness
        flattened_data = {
            **{f"down_{k}": v for k, v in br.down.items()},
            **{f"up_{k}": v for k, v in br.up.items()},
            "down_state": br.down_state,
            "up_state": br.up_state,
            "p": br.p
        }
        data.append(flattened_data)

    # Convert the list of dictionaries to a DataFrame
    df = pd.DataFrame(data)

    # Save the DataFrame as a .parquet file
    df.to_parquet(file_path, engine='pyarrow', index=False)



def load_brs_from_parquet(file_path):
    """
    Load a list of Branch objects from a .parquet file with flattened 'down' and 'up' data.

    Parameters:
    file_path (str): The path to the .parquet file.

    Returns:
    list: A list of Branch objects.
    """
    # Read the DataFrame from the .parquet file
    df = pd.read_parquet(file_path, engine='pyarrow')

    # Reconstruct the list of Branch objects
    brs_list = []
    for _, row in df.iterrows():
        # Reconstruct 'down' and 'up' dictionaries by filtering columns with appropriate prefixes
        down = {key.replace('down_', ''): row[key] for key in row.index if key.startswith('down_') and not key == 'down_state'}
        up = {key.replace('up_', ''): row[key] for key in row.index if key.startswith('up_') and not key == 'up_state'}

        # Create a Branch object with reconstructed dictionaries
        branch = Branch(
            down=down,
            up=up,
            down_state=row.get('down_state'),
            up_state=row.get('up_state'),
            p=row.get('p')
        )
        brs_list.append(branch)

    return brs_list


def get_cmat(branches, comp_varis, sys_st = {'f': 0, 's': 1, 'u': 2}):
    """
    Parameters
    ----------
    branches: list of Branch
    comp_varis: dictionary of variables of component events

    Returns
    C: C matrix of the system event
        The first column is the system state (0: failure, 1: survival, 2: unknown)
    """
    assert isinstance(branches, list), 'branches must be a list'
    assert isinstance(comp_varis, dict), 'comp_var must be a dictionary'

    no_comp = len(comp_varis)

    C = np.zeros((0, no_comp + 1))

    for br in branches:

        c = np.zeros(no_comp + 1)
        c_comp = get_crow(br, comp_varis)
        c[1:] = c_comp

        if br.down_state == 's': # survival branch
            c[0] = sys_st['s']
        elif br.up_state == 'f': # failure branch
            c[0] = sys_st['f']
        else: # unknown branch
            c[0] = sys_st['u']

        C = np.vstack((C, c))
        C = C.astype(int)

    return C


def get_crow(br1, comp_varis):

    c_comp = np.zeros((1, len(comp_varis)))

    for j, (k, v) in enumerate(comp_varis.items()):
        down = br1.down[k]
        up = br1.up[k]

        if up != down:
            bj = {x for x in range(int(down), int(up + 1))}
            sj = v.get_state(bj)
            c_comp[0][j] = sj

        else:
            c_comp[0][j] = up

    return c_comp


'''
def run_bnb(sys_fn, next_comp_fn, next_state_fn, info, comp_max_states):
    """
    return branch
    Parameters
    ----------
    sys_fn
    next_comp_fn
    next_state_fn
    info
    comp_max_states: list-like
    """
    assert callable(sys_fn), 'sys_fn should be a function'
    assert callable(next_comp_fn), 'next_comp_fn should be a function'
    assert callable(next_state_fn), 'next_state_fn should be a function'
    assert isinstance(info, dict), 'info should be a dict'
    assert isinstance(comp_max_states, list), 'comp_max_states should be a list'

    init_up = comp_max_states
    init_down = np.ones_like(comp_max_states).tolist() # Assume that the lowest state is 1

    branches = [Branch_old(down=init_down,
                       up=init_up,
                       is_complete=False,
                       names=info['arcs'])]

    incmp_br_idx = get_idx(branches, False)

    while incmp_br_idx:
        branch_id = incmp_br_idx[0]
        _branch = branches[branch_id]
        down = _branch.down
        up = _branch.up

        down_state, down_val, down_res = sys_fn(down, info)
        up_state, up_val, up_res = sys_fn(up, info)

        if down_state == up_state:
            _branch.is_complete = True
            _branch.down_state = down_state
            _branch.up_state = up_state
            _branch.down_val = down_val
            _branch.up_val = up_val
            del incmp_br_idx[0]

        else:
            # matlab vs python index or not
            #FIXME
            cand_next_comp = [info['arcs'][i] for i, (x, y) in enumerate(zip(up, down)) if x > y]
            #cand_next_comp = [i+1 for i, (x, y) in enumerate(zip(up, down)) if x > y]

            next_comp = next_comp_fn(cand_next_comp, down_res, up_res, info)
            #FIXME
            next_comp_idx = info['arcs'].index(next_comp)
            next_state = next_state_fn(next_comp,
                                       [_branch.down[next_comp_idx], _branch.up[next_comp_idx]],
                                       down_res,
                                       up_res,
                                       info)
            branch_down = copy.deepcopy(_branch)
            branch_down.up[next_comp_idx] = next_state

            branch_up = copy.deepcopy(_branch)
            branch_up.down[next_comp_idx] = next_state + 1

            del branches[branch_id]

            branches.append(branch_down)
            branches.append(branch_up)

            incmp_br_idx = get_idx(branches, False)

    return branches
'''

def get_sb_saved_from_job(output_path, key):
    '''
    output_path: str or Path
    '''
    try:
        assert output_path.exists(), f'output_path does not exist'
    except AttributeError:
        output_path = Path(output_path)
        assert output_path.exists(), f'output_path does not exist'
    finally:
        # read sb_saved json
        sb_saved = []
        for x in output_path.glob(f'sb_dump_{key}_*.json'):
            with open(x, 'r') as fid:
                tmp = json.load(fid)
                [sb_saved.append(tuple(x)) for x in tmp if x[2] == x[3]]

    return sb_saved


def get_bstars_from_sb_dump(file_name):

    bstars = []
    with open(file_name, 'r') as fid:
        tmp = json.load(fid)
        [bstars.append(tuple(x)) for x in tmp if x[2] != x[3]]

    return bstars


def get_sb_saved_from_sb_dump(file_name):

    sb_saved = []
    with open(file_name, 'r') as fid:
        tmp = json.load(fid)
        [sb_saved.append(tuple(x)) for x in tmp if x[2] == x[3]]

    return sb_saved


def split(list_a, chunk_size):

  for i in range(0, len(list_a), chunk_size):
    yield list_a[i:i + chunk_size]


def branch_and_bound(bstars, path_time_idx, arc_cond, output_path, key):
    """
    bstars:
    path_time_idx:
    g_arc_cond:
    """

    i=0
    while bstars:

        print(f'{i}, b*: {len(bstars)}')

        tic = time.perf_counter()

        results = bnb_core(bstars, path_time_idx, arc_cond)

        bstars = [x for result in results for x in result if x not in bstars]

        output_file = output_path.joinpath(f'sb_dump_{key}_{i}.json')
        #with open(output_file, 'w') as w:
        #    json.dump(bstars, w, indent=4)

        print(f'elapsed {i}: {time.perf_counter()-tic:.5f}')

        # next iteration
        bstars = [x for x in bstars if x.up_state != x.down_state]
        i += 1


def bnb_core(bstars, path_time_idx, arc_cond):
    """
    return a list of set of branches given bstars
    path_time_idx: a list of tuples consisting of path, time, and index (corresponding to row of B matrix)
    g_arc_cond: arc_cond dask global variable
    """

    results = []
    for bstar in bstars:
        arcs = get_arcs_given_bstar(bstar, path_time_idx, arc_cond)
        result = get_set_of_branches(bstar, arcs, path_time_idx, arc_cond)
        results.append(result)

    return results


def get_arcs_given_bstar(bstar, path_time_idx, arc_cond):

    lower = bstar.down
    upper = bstar.up

    upper_matched = [k for k, v in upper.items() if v == arc_cond]

    arcs = []
    for _path, _, _ in path_time_idx[1:]:

        if set(_path).issubset(upper_matched):

            arcs = [x for x in _path if upper[x] > lower[x]]

            break

    return arcs


def get_arcs_given_bstar_nobreak(bstar, path_time_idx, arc_cond):

    lower = bstar.down
    upper = bstar.up

    upper_matched = [k for k, v in upper.items() if v == arc_cond]

    _path = next((x for x, _, _ in path_time_idx[1:] if set(x).issubset(upper_matched)), None)

    arcs = [x for x in _path if upper[x] > lower[x]]

    return arcs


def create_arc_state_given_cond(arc, **kwargs):
    """
    return dict of arc state given condition
    arc: str
    kwargs['arc_state']: base arc_state
    kwargs['value']: 0 or 1
    """

    return {k:kwargs['value'] if k==arc else v for k, v in kwargs['arc_state'].items()}


def get_set_of_branches(bstar, arcs, path_time_idx, arc_cond):

    lower = bstar.down
    upper = bstar.up
    c_fl = bstar.down_state
    c_fu = bstar.up_state

    upper_f = {k: 1 if k in arcs else v for k, v in upper.items()}

    sb = []
    for arc in arcs:

        # set upper_n = 0
        #upper = {k: 0 if k == arc else v for k, v in upper.items()}
        upper = create_arc_state_given_cond(arc, value=0, arc_state=bstar.up)

        fu = trans.eval_sys_state_given_arc(upper, path_time_idx=path_time_idx, arc_cond=arc_cond)

        #sb.append((lower, upper, c_fl, fu))
        sb.append(Branch(lower, upper, c_fl, fu))

        # set upper_n=1, lower_n = 1 
        lower = create_arc_state_given_cond(arc, value=1, arc_state=lower)
        #lower = {k: 1 if k == arc else v for k, v in lower.items()}

    #sb.append((lower, upper_f, c_fu, c_fu))
    sb.append(Branch(lower, upper_f, c_fu, c_fu))

    return sb

def branch_and_bound_org(bstars, path_time_idx, arc_cond):
    """
    path_time_idx: a list of tuples consisting of path, time, and index (corresponding to row of B matrix)
    lower:
    upper:
    arc_cond:
    """

    #fl = trans.eval_sys_state(path_time_idx, arcs_state=lower, arc_cond=1)
    #fu = trans.eval_sys_state(path_time_idx, arcs_state=upper, arc_cond=1)

    #sb = [(lower, upper, fl, fu)]

    # selecting a branch from sb such that fl /= fu
    #bstars = [x for x in sb if x[2] != x[3]]

    # make sure the paths are sorted by shortest
    #paths_avail = [x[0] for x in path_time_idx if x[0]]
    sb_saved = []
    while bstars:

        print(f'b*: {len(bstars)}')
        # select path using upper branch of bstar
        for bstar in bstars:

            c_lower, c_upper, c_fl, c_fu = bstar
            upper_matched = [k for k, v in c_upper.items() if v == arc_cond]

            for _path, _, _ in path_time_idx[1:]:

                if set(_path).issubset(upper_matched):

                    upper = c_upper
                    lower = c_lower
                    fl = c_fl
                    chosen = (c_lower, c_upper, c_fl, c_fu)
                    sb = [x for x in bstars if not x == chosen]

                    #paths_avail.remove(_path)
                    break

            for arc in _path:

                if c_upper[arc] > c_lower[arc]:

                    # set upper_n = 0
                    upper = {k: 0 if k == arc else v for k, v in upper.items()}
                    #upper = copy.deepcopy(upper)
                    #upper[arc] = 0
                    fu = trans.eval_sys_state(path_time_idx, upper, arc_cond)
                    #fl = trans.eval_sys_state(path_time_idx, lower, arc_cond)
                    # print(f'{fl} vs {c_fl}')

                    sb.append((lower, upper, fl, fu))

                    # set upper_n=1, lower_n = 1 
                    #upper = c_upper
                    upper = {k: 1 if k == arc else v for k, v in upper.items()}
                    lower = {k: 1 if k == arc else v for k, v in lower.items()}
                    #upper = copy.deepcopy(upper)
                    #upper[arc] = 1
                    #lower = copy.deepcopy(lower)
                    #lower[arc] = 1

            #fu = c_fu
            #fl = trans.eval_sys_state(path_time_idx, lower, arc_cond)

            #if fl==fu:
                #sb.append((lower, upper, fl, fu))
            sb.append((lower, upper, c_fu, c_fu))

        bstars = [x for x in sb if x[2] != x[3]]
        [sb_saved.append(x) for x in sb if x[2] == x[3]]

    return sb_saved

def mcs_unknown_using_cpms(brs, probs, sys_fun, cpms, sys_name, cov_t, rand_seed=None):
    """
    Perform Monte Carlo simulation for the unknown state.

    INPUTS:
    brs_u: Unspecified branches (list)
    probs: a dictionary of failure probabilities for each component
    sys_fun: System function
    cpms: a list of cpms containing component events and system event
    sys_name: a string of the system event's name in cpms
    cov_t: a target c.o.v.
    rand_seed: Random seed

    OUTPUTS:
    cpms: cpms with samples added
    result: Results of the Monte Carlo simulation
    """

    brs_u = [b for b in brs if b.down_state == 'u' or b.up_state == 'u']
    pf_low = sum(b.p for b in brs if b.up_state == 'f')

    # Set the random seed
    if rand_seed:
        np.random.seed(rand_seed)

    brs_u_probs = [b.p for b in brs_u]
    brs_u_prob = sum(brs_u_probs)

    samples = []
    samples_sys = np.empty((0, 1), dtype=int)
    sample_probs = []

    nsamp, nfail = 0, 0
    pf, cov = 0.0, 1.0

    while cov > cov_t:

        nsamp += 1

        sample1 = {}
        s_prob1 = {}

        # select a branch
        br_id = np.random.choice(range(len(brs_u)), p=[p / brs_u_prob for p in brs_u_probs])
        br = brs_u[br_id]

        for e in br.down.keys():
            d = br.down[e]
            u = br.up[e]

            if d < u: # (fail, surv)
                st = np.random.choice(range(d, u + 1), p=[probs[e][d], probs[e][u]])
            else:
                st = d

            sample1[e] = st
            s_prob1[e] = probs[e][st]

        # system function run
        val, sys_st, _ = sys_fun(sample1)

        samples.append(sample1)
        sample_probs.append(s_prob1)
        if sys_st == 'f':
            samples_sys = np.vstack((samples_sys, [0]))
        else: # sys_st == 's'
            samples_sys = np.vstack((samples_sys, [1]))

        if sys_st == 'f':
            nfail += 1

        if nsamp > 9:
            prior = 0.01
            a, b = prior + nfail, prior + (nsamp-nfail) # Bayesian estimation assuming beta conjucate distribution

            pf_s = a / (a+b)
            var_s = a*b / (a+b)**2 / (a+b+1)
            std_s = np.sqrt(var_s)

            pf = pf_low + brs_u_prob *pf_s
            std = brs_u_prob * std_s
            cov = std/pf

            conf_p = 0.95 # confidence interval
            low = beta.ppf(0.5*(1-conf_p), a, b)
            up = beta.ppf(1 - 0.5*(1-conf_p), a, b)
            cint = pf_low + brs_u_prob * np.array([low, up])

        if nsamp % 1000 == 0:
            print(f'nsamp: {nsamp}, pf: {pf:.4e}, cov: {cov:.4e}')

    # Allocate samples to CPMs
    Csys = np.zeros((nsamp, len(probs)), dtype=int)
    Csys = np.hstack((samples_sys, Csys))

    for i, v in enumerate(cpms[sys_name].variables[1:]):
        Cv = np.array([s[v.name] for s in samples], dtype=int).T
        cpms[v.name].Cs = Cv
        cpms[v.name].q = np.array([p[v.name] for p in sample_probs], dtype=float).T
        cpms[v.name].sample_idx = np.arange(nsamp, dtype=int)

        Csys[:, i+1] = Cv.flatten()

    cpms[sys_name].Cs = Csys
    cpms[sys_name].q = np.ones((nsamp,1), dtype=float)
    cpms[sys_name].sample_idx = np.arange(nsamp, dtype=int)

    result = {'pf': pf, 'cov': cov, 'nsamp': nsamp, 'cint_low': cint[0], 'cint_up': cint[1]}

    return cpms, result

def mcs_unknown_simple(brs, probs, sys_fun, cov_t, rand_seed=None):
    """
    Perform Monte Carlo simulation for the unknown state WITHOUT involving CPMs.
    (used when re-use of MCS samples is not considered)

    INPUTS:
    brs_u: Unspecified branches (list)
    probs: a dictionary of failure probabilities for each component
    sys_fun: System function
    cov_t: a target c.o.v.
    rand_seed: Random seed

    OUTPUTS:
    cpms: cpms with samples added
    result: Results of the Monte Carlo simulation
    """

    brs_u = [b for b in brs if b.down_state == 'u' or b.up_state == 'u']
    pf_low = sum(b.p for b in brs if b.up_state == 'f')

    # Set the random seed
    if rand_seed:
        np.random.seed(rand_seed)

    brs_u_probs = [b.p for b in brs_u]
    brs_u_prob = sum(brs_u_probs)

    nsamp, nfail = 0, 0
    pf, cov = 0.0, 1.0

    while cov > cov_t:

        nsamp += 1
        sample1 = {}

        # select a branch
        br_id = np.random.choice(range(len(brs_u)), p=[p / brs_u_prob for p in brs_u_probs])
        br = brs_u[br_id]

        for e in br.down.keys():
            d = br.down[e]
            u = br.up[e]

            if d < u: # (fail, surv)
                st = np.random.choice(range(d, u + 1), p=[probs[e][d], probs[e][u]])
            else:
                st = d

            sample1[e] = st

        # system function run
        val, sys_st, _ = sys_fun(sample1)

        if sys_st == 'f':
            nfail += 1

        if nsamp > 9:
            prior = 0.01
            a, b = prior + nfail, prior + (nsamp-nfail) # Bayesian estimation assuming beta conjucate distribution

            pf_s = a / (a+b)
            var_s = a*b / (a+b)**2 / (a+b+1)
            std_s = np.sqrt(var_s)

            pf = pf_low + brs_u_prob *pf_s
            std = brs_u_prob * std_s
            cov = std/pf

            conf_p = 0.95 # confidence interval
            low = beta.ppf(0.5*(1-conf_p), a, b)
            up = beta.ppf(1 - 0.5*(1-conf_p), a, b)
            cint = pf_low + brs_u_prob * np.array([low, up])

        if nsamp % 1000 == 0:
            print(f'nsamp: {nsamp}, pf: {pf:.4e}, cov: {cov:.4e}')


    result = {'pf': pf, 'cov': cov, 'nsamp': nsamp, 'cint_low': cint[0], 'cint_up': cint[1]}

    return result
